from PyQt5.QtWidgets import QVBoxLayout, QLabel, QPushButton, QHBoxLayout, QWizard, QWizardPage, QSpinBox, QComboBox, QCheckBox
from PyQt5.QtCore import pyqtSignal
from yeastvision.parts.dialogs import InterpolationIntervalWidget
from yeastvision.parts.guiparts import NumericLineEdit


class SimpleTextPage(QWizardPage):
    def __init__(self, text, parent=None):
        super().__init__(parent)
        self.setTitle("Information Page")
        
        layout = QVBoxLayout()
        text_label = QLabel(text)
        layout.addWidget(text_label)
        self.setLayout(layout)


class InterpolationDialog(QWizardPage):
    completeChanged = pyqtSignal()
    def __init__(self, channel_options, text = None, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.setTitle("Select Interpolation Intervals")
        
        self.layout = QVBoxLayout()

        if text:
            self.layout.addWidget(text)

        self.error_label = QLabel("")
        self.error_label.setStyleSheet("color: red")
        self.layout.addWidget(self.error_label)
        
        self.add_interval_button = QPushButton("Add Interval")
        self.add_interval_button.clicked.connect(self.add_interval)
        
        self.layout.addWidget(self.add_interval_button)
        self.setLayout(self.layout)
        
        self.intervals = []
        self.channels = channel_options
        self.channel_obj = None
    
    def initializePage(self):
        self.remove_all_intervals()
        channel_index = self.field("channelIndex")
        self.channel_obj = self.channels[channel_index]
        self.update_max_frame()

    def isComplete(self):
        # Page is complete if a valid channel is selected (index >= 0)
        return self.check_intervals()

    def remove_all_intervals(self):
        for interval_widget in self.intervals:
            interval_widget.setParent(None)
        self.intervals.clear()

    def disable_and_clear_interval_add(self):
        self.remove_all_intervals()
        self.add_interval_button.setDisabled(True)
    
    def enable_interval_add(self):
        self.add_interval_button.setDisabled(False)

    def add_interval(self):
        max_frame = self.get_max_frame()
        interval_widget = InterpolationIntervalWidget(max_frame=max_frame)
        interval_widget.removed.connect(self.remove_interval)  # Connect the signal
        self.intervals.append(interval_widget)
        self.layout.insertWidget(self.layout.count() - 1, interval_widget)
        self.completeChanged.emit()

    def remove_interval(self, interval_widget):
        if interval_widget in self.intervals:
            self.intervals.remove(interval_widget)
            interval_widget.deleteLater()
            self.completeChanged.emit()

    def update_max_frame(self):
        max_frame = self.get_max_frame()
        if max_frame == 0:
            self.disable_and_clear_interval_add()
            self.showMessage("Current Channel has only one image.")
        else:
            self.add_interval_button.setDisabled(False)
            self.clearMessages()
            for interval_widget in self.intervals:
                interval_widget.set_max_frame(max_frame)

    def get_max_frame(self):
        return self.channel_obj.max_t()

    def check_intervals(self):
        new_intervals = []
        for interval_widget in self.intervals:
            start_frame = interval_widget.start_frame_spinbox.value()
            end_frame = interval_widget.end_frame_spinbox.value()
            interp = interval_widget.interpolation_combo.currentText()
            max_frame = self.get_max_frame()

            if start_frame < 0 or start_frame >= max_frame:
                self.showMessage(f"Start frame {start_frame} is out of range.")
                return False
            if end_frame <= start_frame or end_frame > max_frame:
                self.showMessage(f"End frame {end_frame} is out of range or not greater than start frame.")
                return False

            new_intervals.append((start_frame, end_frame, interp))

        # Check for overlapping intervals
        new_intervals.sort()  # Sort intervals by start frame
        for i in range(len(new_intervals) - 1):
            if new_intervals[i][1] > new_intervals[i + 1][0]:
                self.showMessage("Intervals overlap.")
                return False

        self.clearMessages()
        return True
    
    def getData(self):
        intervals = [interval_widget.get_interval_data() for interval_widget in self.intervals]
        intervals = sorted(intervals, key=lambda x: x['start'])
        return {"intervals": intervals}

    def showMessage(self, message):
        self.error_label.setText(message)

    def clearMessages(self):
        self.showMessage("") 


class ParameterInputPage(QWizardPage):
    def __init__(self, params_dict, model_weights, modelName, channels, parent=None, includeTime=True):
        super().__init__(parent)
        self.setTitle(f"Set {modelName} Parameters")
        self.modelName = modelName
        self.params_dict = params_dict
        self.includeTime=includeTime

        # Layout for the parameters
        self.param_layout = QVBoxLayout()

        # Create QLineEdits for each parameter
        self.line_edits = {}
        self.check_boxes = {}
        for key, value in params_dict.items():
            h_layout = QHBoxLayout()
            label = QLabel(f"{key}:")
            if isinstance(value, bool):
                checkbox = QCheckBox(self)
                checkbox.setCheckState(value)
                self.check_boxes[key] = checkbox
                h_layout.addWidget(label)
                h_layout.addWidget(checkbox)
                self.param_layout.addLayout(h_layout)
            else:
                line_edit = NumericLineEdit(str(value))
                self.line_edits[key] = line_edit
                h_layout.addWidget(label)
                h_layout.addWidget(line_edit)
                self.param_layout.addLayout(h_layout)

        if self.includeTime:
            # Create spin boxes for t_start and t_stop
            self.t_start_spinbox = QSpinBox()
            self.t_stop_spinbox = QSpinBox()

            self.t_start_spinbox.setRange(0, 1000)  # Arbitrary large max, will be set correctly in initializePage
            self.t_stop_spinbox.setRange(0, 1000)

            # Create layout for the spin boxes
            spin_layout = QHBoxLayout()
            spin_layout.addWidget(QLabel("t_start:"))
            spin_layout.addWidget(self.t_start_spinbox)
            spin_layout.addWidget(QLabel("t_stop:"))
            spin_layout.addWidget(self.t_stop_spinbox)

        # Create combo box for model weights
        self.model_weights_combo = QComboBox()
        self.model_weights_combo.addItems(model_weights)
        self.model_weights_combo.setCurrentIndex(0)
        model_weights_layout = QHBoxLayout()
        model_weights_layout.addWidget(QLabel("Choose model weights:"))
        model_weights_layout.addWidget(self.model_weights_combo)

        # Main layout
        layout = QVBoxLayout()
        layout.addLayout(self.param_layout)
        if self.includeTime:
            layout.addLayout(spin_layout)
        layout.addLayout(model_weights_layout)
        self.setLayout(layout)

        # Connect signals to ensure proper min/max behavior
        if self.includeTime:
            self.t_start_spinbox.valueChanged.connect(self.update_t_stop_min)
            self.t_stop_spinbox.valueChanged.connect(self.update_t_start_max)

        self.channels = channels
        self.channel_obj = None

    def update_t_stop_min(self, value):
        self.t_stop_spinbox.setMinimum(value)

    def update_t_start_max(self, value):
        self.t_start_spinbox.setMaximum(value)

    def initializePage(self):
        # Retrieve the channel index
        channel_index = self.field("channelIndex")
        self.channel_obj = self.channels[channel_index]
        max_t = self.channel_obj.max_t()+1
        
        if self.includeTime:
            self.t_start_spinbox.setMaximum(max_t)
            self.t_stop_spinbox.setMaximum(max_t)
            self.t_stop_spinbox.setValue(max_t)
        
    def getData(self):
        data = {
            "modelName": self.modelName,
            "modelWeight": self.model_weights_combo.currentText(),
        }
        
        if self.includeTime:
            data.update({"t_start": int(self.t_start_spinbox.value()), 
                         "t_stop": int(self.t_stop_spinbox.value())}
            )
        for key in self.line_edits:
            data[key] = float(self.line_edits[key].text())
        for key in self.check_boxes:
            data[key] = bool(self.check_boxes[key].isChecked())
        return data




import numpy as np
from skimage.measure import regionprops_table
import pyqtgraph as pg
from PyQt5 import QtCore
'''
Plots types are divided into two large categories

Single Cell Plots - defined by cellDataDf (cell select change)
    Image Plots - mean, var, median vs time
    Morphological Plots - cell properties vs time
    Space plots - y vs x


Population Plots (label change)
    Average Image (remove zeros)
    Average Morphology (remove zeros)
    Count - vs time: option for curve fit
    Heatmaps (clustered)

Plot format
--Single Cell--                       ----Population Plots----
                                          Count vs Time 
Intensity Plots - tabbed                  Mean Intensity
Morphogical Plots - tabbed                Mean Morphologies
Space plot                                Heatmaps


Option to add a plot


Option to add a plot:                                                             New Plots
Property Plot
    Add single cell plot [] -->  Cell Number [dropdown] Property [dropdown]       -------------
    Add Population Plot [] -->  Plot Type                                         -------------

Correlation Plot
'''

def yvv(celldata):
    pass

class SingleFrameUpdatePlot:
    def __init__(self, parent, property_name, mask_name, data):
        self.parent = parent
        self.property_name = property_name
        self.mask_name = mask_name
        self.data = data
        self.plot = pg.PlotWidget(title=f"{self.property_name} for {self.mask_name}")
    
    def getData(self, frame_index=None):
        if frame_index is not None and isinstance(self.data, list):
            if frame_index < len(self.data):
                frame_data = self.data[frame_index]
                if self.property_name in frame_data:
                    return np.array(frame_data[self.property_name])
        elif self.property_name in self.data:
            return np.array(self.data[self.property_name])
        return np.array([])
    
    def update(self, frame_index=None):
        self.plot.clear()
        data = self.getData(frame_index)
        if data.size > 0:
            y, x = np.histogram(data, bins="auto")
            self.plot.plot(x, y, stepMode=True, fillLevel=0, brush=(0,0,255,150))
    
    def getPlotWidget(self):
        self.update()
        return self.plot

class SingleCellUpdatePlot():
    def __init__(self, parent, property, timeseries, selectedCells):
        self.parent = parent
        self.property = property
        self.timeseries = timeseries
        self.tsNames = [self.parent.getTimeSeriesDataName(ts) for ts in self.timeseries]
        self.selectedCells = selectedCells
        self.plot = pg.PlotWidget(title=self.property)
    
    def getData(self):
        data = {}
        if self.selectedCells:
            for cell in self.selectedCells:
                for ts, name in zip(self.timeseries, self.tsNames):
                    if cell in ts.label_vals:
                        data[f"{name}-cell{cell}"] = ts.get(self.property, cell)
        return data
    
    def update(self):
        self.plot.clear()
        data = self.getData()
        if data:
            if len(data) > 1:
                self.plot.addLegend()
    
            i = 0
            for cellnum, dataset in data.items():
                color = self.parent.cmap[i][0:-1]
                self.plot.plot(dataset, name = str(cellnum), pen = tuple(color))
                i+=1
            
    def getPlotWidget(self):
        self.update()
        return self.plot


class PopulationPlot():
    def __init__(self, parent, property, timeseries, populationName):
        self.parent = parent
        self.property = property
        self.data = timeseries
        self.population = populationName
        self.plot = pg.PlotWidget(title=f"{self.property}-{self.population}")
    
    def getData(self):
        return self.data.get_population_data(self.property, self.population)

    def getVar(self):
        return self.data.get_population_data(self.property, self.population, func = np.std)
    
    def update(self):
        self.plot.clear()
        data = self.getData()
        variance = self.getVar()

        plot = self.plot

        x = np.arange(len(data))
        # Plot the data with a bold line
        plot.plot(data, pen='b', width=2, name=f"{self.property}-mean")

        # Plot the data points plus or minus the variance
        plot.plot(data + variance, pen='r', style=QtCore.Qt.DashLine, width=1)
        plot.plot(data - variance, pen='r', style=QtCore.Qt.DashLine, width=1)

        # Add error bars at each time point
        errorbar = pg.ErrorBarItem(x=x, y=data, height=variance*2, beam=0.2)
        plot.addItem(errorbar)

        # Shade the region between the plus or minus variance datapoints
        fill = pg.FillBetweenItem(curve1=plot.getPlotItem().listDataItems()[1],
                                    curve2=plot.getPlotItem().listDataItems()[-1],
                                    brush=(100, 100, 100, 100))
        plot.addItem(fill)

        # Set the x-axis and y-axis labels
        plot.setLabel('bottom', 'Time')
        plot.setLabel('left', 'Value')

        # Add a legend
        plot.addLegend()

    def getPlotWidget(self):
        self.update()
        return self.plot


class HeatMap():
    def __init__(self, parent, prop, populationName, data):
        self.parent = parent
        self.data = data
        self.populationName = populationName
        self.property = prop
        self.cmap = pg.colormap.get("jet", source = 'matplotlib')
        self.view  = pg.PlotWidget(title = f"{self.populationName}-{self.property}")
        self.im = pg.ImageItem(colorMap = self.cmap)
        self.bar = pg.ColorBarItem( values=(0,1), colorMap=self.cmap)
        self.bar.setImageItem(self.im, insert_in = self.view.getPlotItem())
        self.view.addItem(self.im)
    
    def getData(self):
        return self.data.get_heatmap_data(self.property, self.populationName)
    
    def update(self):
        imData = self.getData()
        self.im.setImage(imData, levels = [0,1])
    
    def getPlotWidget(self):
        self.update()
        return self.view



from yeastvision.models.cp import CustomCPWrapper

class ProSeg(CustomCPWrapper):
    hyperparams  = { 
    "mean_diameter":30.0,
    "flow_threshold":0.9, 
    "cell_probability_threshold": 0.5}
    types = [float, float,float]

    def __init__(self, params, weights):
        super().__init__(params, weights)

# # need to update here
# class ArtilifeFullLifeCycle(ProSeg):
#     hyperparams  = { 
#     "mean_diameter":30, 
#     "flow_threshold":0.9, 
#     "cell_probability_threshold": 0.5,
#     "has_mating": False,
#     "has_spores": False,
#     "is_time_series": True}
#     types = [None, None,None, bool, bool, bool, bool]

#     '''
#     params also include:
#         matStart - start index of mating cells
#         matStop - stop index of mating cells (inclusive)
#         matSeg - weights to use for matSeg 

#         sporeStart - start index of sporulating cells
#         sporeStop - stop index of sporulating cells (inclusive)
#         tetradSeg - weights to use for tetradSeg
#     these are hard coded via guiparts.ArtilifeParamDialog
#     '''
#     def __init__(self, params, weights):
#         super().__init__(params, weights)

#         self.matSeg, self.tetradSeg = None,None
#         self.matMasks, self.matprobs = None, None
#         self.tetraMasks, self.tetraprobs = None, None

#         if params["has_mating"]:
#             from models.matSeg.model import MatSeg
#             self.matSeg = MatSeg(self.params, produce_weight_path("matSeg", self.params["matSeg"]))

#         if params["has_spores"]:
#             from models.spoSeg.model import TetradSeg
#             self.tetradSeg = TetradSeg(self.params, produce_weight_path("tetradSeg", self.params["tetradSeg"]))

#     def addTetrads(self, ims):
#         tetraSlice = slice(int(self.params["sporeStart"]), int(self.params["sporeStop"])+1)
#         self.tetraMasks, tetraFlows, _, _ = self.tetradSeg.model.eval(ims[tetraSlice], 
#                                                                 diameter = self.params["Mean Diameter"], 
#                                                                 channels = [0,0],
#                                                                 cellprob_threshold = self.params["Flow Threshold"],
#                                                                 do_3D = False)
#         self.tetraprobs = [flow[2] for flow in tetraFlows]
#         self.tetraprobs = np.array(self.process_probability(self.tetraprobs), dtype = np.uint8)

#         if self.params["is_time_series"]:
#             newTetraMasks, newMasks = track_obj([im[:,:,0] for im in ims[tetraSlice]], self.masks[tetraSlice], self.tetraMasks[tetraSlice], False)
#             tetraSlice = slice(int(self.params["sporeStart"]), int(self.params["sporeStop"])-1)
#         else:
#             for tetraMask, cellMask in zip(self.tetraMasks, self.masks):
#                 if np.any(tetraMask>=1):
#                     newMask = addMasks(tetraMask, cellMask)
#                     newMasks.append(newMask)
#                 else:
#                     newMasks.append(cellMask)
#             newTetraMasks = self.tetraMasks

#         self.masks[tetraSlice] = newMasks
#         tetraMasks = np.zeros_like(self.masks)
#         tetraMasks[tetraSlice] = np.array(newTetraMasks, dtype = np.uint16)
#         self.tetraMasks = tetraMasks


#     def addMatingCells(self, ims):
#         matSlice = slice(int(self.params["matStart"]),int(self.params["matStop"]) + 1)
#         self.matMasks, matFlows, _, _ = self.matSeg.model.eval(ims[matSlice], 
#                                                                 diameter = self.params["mean_diameter"], 
#                                                                 channels = [0,0],
#                                                                 cellprob_threshold = self.params["flow_threshold"],
#                                                                 do_3D = False)
#         print("got mating cells")
#         self.matprobs = [flow[2] for flow in matFlows]
#         self.matprobs = np.array(self.process_probability(self.matprobs), dtype = np.uint8)
        
#         if self.params["is_time_series"]:
#             matSlice = slice(int(self.params["matStart"]), int(self.params["matStop"])-1)
#             newMatMasks, newMasks = get_mating_data(self.matMasks[matSlice], self.masks[matSlice])
#         else:
#             newMasks = []
#             for matMask, cellMask in zip(self.matMasks, self.masks):
#                 if np.any(matMask>=1):
#                     newMask = addMasks(matMask, cellMask)
#                     newMasks.append(newMask)
#                 else:
#                     newMasks.append(cellMask)
#             newMatMasks = self.matMasks

#         self.masks[matSlice] = newMasks
#         matMasks = np.zeros_like(self.masks)
#         matMasks[matSlice] = np.array(newMatMasks, dtype = np.uint16)
#         self.matMasks = matMasks

#     @classmethod
#     @torch.no_grad()
#     def run(cls, ims, params, weights):
#         params = params if params else cls.hyperparams
#         model = cls(params, weights)
#         ims3D = [cv2.merge((im,im,im)) for im in ims]
#         assert len(ims3D[0].shape)==3

#         if not params["Mean Diameter"]:
#             evaluator = model.cpAlone
#             model.masks, flows, _ = evaluator.eval(ims3D, 
#                                                     diameter = model.params["Mean Diameter"], 
#                                                     channels = [0, 0],
#                                                     cellprob_threshold = model.params["Flow Threshold"], 
#                                                     do_3D=False)
#         else:
#             evaluator = model.model
#             model.masks, flows, _, model.diams = evaluator.eval(ims3D, 
#                                                     diameter = model.params["Mean Diameter"], 
#                                                     channels = [0, 0],
#                                                     cellprob_threshold = model.params["Flow Threshold"], 
#                                                     do_3D=False)
#         print("process probability")
#         model.cellprobs = [flow[2] for flow in flows]
#         model.cellprobs = np.array((model.process_probability(model.cellprobs)), dtype = np.uint8)
        
#         print("formatting return")
#         model.masks = np.array(model.masks, dtype = np.uint16)

#         if model.matSeg:
#             print('add mating cell')
#             model.addMatingCells(ims3D)
#         if model.tetradSeg:
#             model.addTetrads(ims3D)
        
#         print("finished")
#         arti = (model.masks, model.cellprobs)
#         mating = (model.matMasks, model.matprobs)
#         tetra = (model.tetraMasks, model.tetraprobs)

#         del model

#         return {"artilife": arti,
#                 "mating": mating,
#                 "tetrads": tetra
#                 }


    



import os
import glob
import numpy as np
import uuid
import hashlib

image_extensions = ['.tif','.tiff',
                            '.jpg','.jpeg','.png','.bmp',
                            '.pbm','.pgm','.ppm','.pxm','.pnm','.jp2',
                            '.TIF','.TIFF',
                            '.JPG','.JPEG','.PNG','.BMP',
                            '.PBM','.PGM','.PPM','.PXM','.PNM','.JP2']


def get_id_from_name(name: str) -> str:
    """
    Generates a unique ID based on the provided string.
    
    Args:
        name (str): The input string for which a unique ID is needed.
        
    Returns:
        str: A unique hexadecimal ID corresponding to the input string.
    """
    # Create a unique hash using SHA-256
    hash_object = hashlib.sha256(name.encode())
    unique_id = hash_object.hexdigest()
    
    return unique_id

def get_id() -> str:
    """
    Generates a unique ID that is not based on any input string.
    
    Returns:
        str: A unique ID.
    """
    # Generate a random UUID
    unique_id = str(uuid.uuid4())
    
    return unique_id

def get_file_name(path):
    _,path = os.path.split(path)
    return path

def get_files(dir, extension):
    return sorted(glob.glob(os.path.join(dir, "*"+extension)))

def has_extension(dir, extension):
    return bool(get_files(dir, extension))

def get_im_mask_npzs(dir):
    npz_files = get_files(dir, ".npz")
    ims, masks = [], []
    if npz_files:
        for file in npz_files:
            if is_mask_npz(file):
                masks.append(file)
            elif is_image_npz(file):
                ims.append(file)
    return ims,masks

def is_mask_npz(file):
    return "labels" in np.load(file).files

def is_image_npz(file):
    return "ims" in np.load(file).files


def get_extensions_in_dir(dir):
    extensions = []
    for path in os.listdir(dir):
        name, ext = os.path.splitext(path)
        if ext not in extensions:
            extensions.append(ext)
    return extensions

def has_label_npz(dir):
    for path in os.listdir(dir):
        name,ext = os.path.splitext(path)
        if name == "labels" and ext == ".npz":
            return True

def get_filetype(path):
    _, ext = os.path.splitext(path)
    return ext

def is_image(path):
    return get_filetype(path) in image_extensions

def all_files(dir):
    return glob.glob(os.path.join(dir, "*"))


def get_image_files(dir, remove_masks = False):
    files = all_files(dir)
    all_ims = [im for im in files if is_image(im)]
    if remove_masks:
        return [im for im in all_ims if "_mask" not in get_file_name(im)]
    else:
        return all_ims

def concatenate_arrays_by_dict(main_dict, new_dict):
    for key in main_dict.keys():
        new_data = new_dict[key]
        main_data = main_dict[key]
        main_dict[key] = np.concatenate((main_data, new_data), axis = 0)
    return main_dict


def append_to_npz(npz_file, new_data):
    npz_dict = dict(np.load(npz_file))
    np.savez(npz_file, **concatenate_arrays_by_dict(npz_dict, new_data))


from scipy.stats import mode
from skimage.morphology import thin
from yeastvision.track.fiest.utils import remove_artif, binar, OAM_23121_tp3
import numpy as np
from typing import List, Optional
import time
import logging

logger = logging.getLogger(__name__)

##### one of the steps of the fiest full lifecycle tracking code
##### TODO: port this to main tracking module
## corresponds to step5 of the pipeline
def track_general_masks(Art_MT, disk_size = 6, transpose_out=False):
    '''
    Tracking alogrithm for budding yeast with built-in track corrections. Iterative frame-assignmentper object asssumes frame-to-frame overlap. For fast
    growing colonies or cells that move significantly between frames, explore using interpolation before calling this function (see yeastvision.track.fiest.track for the FIEST algorithm or 
    yeastvision.ims.interpolate for fine tuned control of interpolation)
    
    Args: 
        Art-MT (np.ndarray or list of np.ndarray) - 
        disk_size (int, default 6)
        
    Returns:
        np.ndarray - '''
    logger.info("----- General tracking of Saccharomyces cerevisiae using mask overlap----")
    Masks3 = Art_MT
        
    im_no1 = 0
    im_no = len(Masks3)
    mm = range(im_no) # time points to track
    
    IS1 = np.copy(Masks3[im_no1]).astype('uint16') # start tracking at first time point # plt.imshow(IS1)
    IS1 = remove_artif(IS1, disk_size) # remove artifacts and start tracking at first time point # plt.imshow(IS1)
    masks = np.zeros((IS1.shape[0], IS1.shape[1], im_no)) # contains the re-labeled masks according to labels in the last tp mask
    masks[:,:,im_no1] = IS1.copy()
 
    """
    Allocate a mask for cells where there's a gap in the segmentation; IblankG is updated within the loops it1 and itG
    """
    IblankG = np.zeros(IS1.shape, dtype="uint16")
    tic = time.time()
    for it0 in mm: # notice IS1 will be updated in the loop it0=0
        #print(f'it0={it0}')
        # Load the future cellpose mask, IS2: IS2 is the current image being re-indexed for tracking
        IS2 = np.copy(Masks3[it0]).astype('uint16') # plt.imshow(IS2)
 
        IS2 = remove_artif(IS2, disk_size) # set disk_size as needed # 5 is ideal to match MATLAB's disk_size=6
        
        IS2C = np.copy(IS2) # plt.imshow(IS2C) # <--- a copy of IS2, gets updated in it1
        IS1B = binar(IS1)
        
        IS3 = IS1B.astype('uint16') * IS2 # past superimposed to future; updated in it1
        tr_cells = np.unique(IS1[IS1 != 0]) # the tracked cells present in the present mask, IS1
        
        gap_cells = np.unique(IblankG[IblankG != 0]) # the tracked cells that had a gap in their segmentation; were not detected in IS1
        cells_tr = np.concatenate((tr_cells, gap_cells)) # all the cells that have been tracked up to this tp for this position

        
        # Allocate space for the re-indexed IS2 according to tracking
        Iblank0 = np.zeros_like(IS1)
        
        # Go to the previously tracked cells and find corresponding index in current tp being processed, IS2 -> Iblank0: mask of previously tracked cells with new position in IS2
        
        if cells_tr.sum() != 0: # this is required in case the mask goes blank because cells mate immediately during germination
            for it1 in np.sort(cells_tr): # cells are processed in order according to birth/appearance
                IS5 = (IS1 == it1).copy() # go to past mask, IS1, to look for the cell
                IS6A = np.uint16(thin(IS5, max_num_iter=1)) * IS3

                if IS5.sum() == 0: # if the cell was missing in the past mask; look at the gaps in segmentation, otherwise, continue to look at the past mask
                    IS5 = (IblankG == it1).copy()
                    IS6A = np.uint16(thin(IS5, max_num_iter=1)) * IS2C
                    IblankG[IblankG == it1] = 0 # remove the cell from the segmentation gap mask - it'll be updated in the past mask for next round of processing

                # Find the tracked cell's corresponding index in IS2, update IS3 and IS2C to avoid overwriting cells 
                if IS6A.sum() != 0:
                    IS2ind = 0 if not IS6A[IS6A != 0].any() else mode(IS6A[IS6A != 0])[0]
                    Iblank0[IS2 == IS2ind] = it1
                    IS3[IS3 == IS2ind] = 0
                    IS2C[IS2 == IS2ind] = 0

            # Define cells with segmentation gap, update IblankG, the segmentation gap mask
            seg_gap = np.setdiff1d(tr_cells, np.unique(Iblank0)) # cells in the past mask (IS1), that were not found in IS2 

            if seg_gap.size > 0:
                for itG in seg_gap:
                    IblankG[IS1 == itG] = itG

            # Define cells that were not relabelled in IS2; these are the buds and new cells entering the frame
            Iblank0B = Iblank0.copy()
            Iblank0B[Iblank0 != 0] = 1
            ISB = IS2 * np.uint16(1 - Iblank0B)
            
            # Add new cells to the mask with a new index Iblank0->Iblank, Iblank0 with new cells added
            newcells = np.unique(ISB[ISB != 0])
            Iblank = Iblank0.copy()
            A = 1

            if newcells.size > 0:
                for it2 in newcells:
                    Iblank[IS2 == it2] = np.max(cells_tr) + A # create new index that hasn't been present in tracking
                    A += 1

            masks[:, :, it0] = np.uint16(Iblank).copy() #<---convert tracked mask to uint16 and store
            IS1 = masks[:, :, it0].copy() # IS1, past mask, is updated for next iteration of it0

        else:
            masks[:, :, it0] = IS2.copy()
            IS1 = IS2.copy()

    toc = time.time()
    logger.info(f'GENERAL TRACKING: Elapsed time is {toc - tic} seconds.')


    """
    Vizualize All Ob
    """
    obj = np.unique(masks)
    no_obj = int(np.max(obj))
    im_no = masks.shape[2]
    all_ob = np.zeros((no_obj, im_no))

    tic = time.time()

    for ccell in range(1, no_obj + 1):
        Maa = (masks == ccell)

        for i in range(im_no):
            pix = np.sum(Maa[:, :, i])
            all_ob[ccell-1, i] = pix

    """
    Tracks as a tensor
    """

    im_no = masks.shape[2]
    # Find all unique non-zero cell identifiers across all time points
    ccell2 = np.unique(masks[masks != 0])
    # Initialize Mask2 with zeros of the same shape as masks
    Mask2 = np.zeros((masks.shape[0], masks.shape[1], masks.shape[2]))


    # instead of np use cpypy?
    # Process each unique cell ID
    for itt3 in range(len(ccell2)):  # cells
        pix3 = np.where(masks == ccell2[itt3])
        Mask2[pix3] = itt3 + 1  # ID starts from 1

    """
    Get Cell Presence
    """

    # Get cell presence
    Mask3 = Mask2.copy()
    numbM = im_no
    obj = np.unique(Mask3)
    no_obj1 = int(obj.max())
    A = 1

    tp_im = np.zeros((no_obj1, im_no))

    for cel in range(1, no_obj1+1):
        Ma = (Mask3 == cel)

        for ih in range(numbM):
            if Ma[:, :, ih].sum() != 0:
                tp_im[cel-1, ih] = 1


    """
    Split Interrupted time series
    """

    tic = time.time()
    for cel in range(1, no_obj1+1):
        #print(cel)
        tp_im2 = np.diff(tp_im[cel-1, :])
        tp1 = np.where(tp_im2 == 1)[0]
        tp2 = np.where(tp_im2 == -1)[0]
        maxp = (Mask3[:, :, numbM - 1] == cel).sum()

        if len(tp1) == 1 and len(tp2) == 1 and maxp != 0:  # has one interruption
            for itx in range(tp1[0], numbM):
                tp3 = OAM_23121_tp3(Mask3[:, :, itx], cel, no_obj1, A)
                Mask3[:, :, itx] = tp3.copy()
            no_obj1 += A
        
        elif len(tp1) == 1 and len(tp2) == 1 and maxp == 0:  # has one interruption
            pass
        
        elif len(tp1) == len(tp2) + 1 and maxp != 0:
            tp2 = np.append(tp2, numbM-1)

            for itb in range(1, len(tp1)):  # starts at 2 because the first cell index remains unchanged
                for itx in range(tp1[itb] + 1, tp2[itb] + 1):
                    tp3 = OAM_23121_tp3(Mask3[:, :, itx], cel, no_obj1, A)
                    Mask3[:, :, itx] = tp3.copy()
                no_obj1 += A
        
        elif len(tp2) == 0 or len(tp1) == 0:  # it's a normal cell, it's born and stays until the end
            pass
        
        elif len(tp1) == len(tp2):
            if tp1[0] > tp2[0]:
                tp2 = np.append(tp2, numbM-1) #check this throughly
                for itb in range(len(tp1)):
                    for itx in range(tp1[itb]+1, tp2[itb + 1] + 1):
                        tp3 = OAM_23121_tp3(Mask3[:, :, itx], cel, no_obj1, A) #+1 here
                        Mask3[:, :, itx] = tp3.copy()    
                    no_obj1 += A
            elif tp1[0] < tp2[0]:
                for itb in range(1, len(tp1)): 
                    for itx in range(tp1[itb] + 1, tp2[itb] + 1):  # Inclusive range
                        tp3 = OAM_23121_tp3(Mask3[:, :, itx], cel, no_obj1, A)
                        Mask3[:, :, itx] = tp3.copy()
                    no_obj1 += A
            elif len(tp2) > 1:
                for itb in range(1, len(tp1)):
                    for itx in range(tp1[itb] + 1, tp2[itb] + 1):
                        tp3 = OAM_23121_tp3(Mask3[:, :, itx], cel, no_obj1, A)
                        Mask3[:, :, itx] = tp3.copy()    
                    no_obj1 += A
    toc = time.time()
    logger.info(f'GENERAL TRACKING: Elapsed time is {toc - tic} seconds.')


    """
    Get cell presence 2
    """
    numbM = im_no
    obj = np.unique(Mask3)

    # Get cell presence 2
    tp_im = np.zeros((int(max(obj)), im_no))

    for cel in range(1, int(max(obj)) + 1):
        Ma = (Mask3 == cel)

        for ih in range(numbM):
            if Ma[:, :, ih].sum() != 0:
                tp_im[cel-1, ih] = 1

    """
    Get good cells

    """
    cell_artifacts = np.zeros(tp_im.shape[0])

    for it05 in range(tp_im.shape[0]):
        arti = np.where(np.diff(tp_im[it05, :]) == -1)[0]  # Find artifacts in the time series

        if arti.size > 0:
            cell_artifacts[it05] = it05 + 1  # Mark cells with artifacts

    goodcells = np.setdiff1d(np.arange(1, tp_im.shape[0] + 1), cell_artifacts[cell_artifacts != 0])  # Identify good cells


    """
    Tracks as a tensor 2
    """
    im_no = Mask3.shape[2]
    Mask4 = np.zeros((masks.shape[0], masks.shape[1], masks.shape[2]))

    for itt3 in range(goodcells.size):
        pix3 = np.where(Mask3 == goodcells[itt3])
        Mask4[pix3] = itt3 + 1  # IDs start from 1


    """
    # Get cell presence 3
    """

    Mask5 = Mask4.copy()
    numbM = im_no
    obj = np.unique(Mask4)
    no_obj1 = int(obj.max())
    A = 1

    tp_im = np.zeros((no_obj1, im_no))

    for cel in range(1, no_obj1+1):
        Ma = (Mask5 == cel)

        for ih in range(numbM):
            if Ma[:, :, ih].sum() != 0:
                tp_im[cel-1, ih] = 1

    #######
    cell_exists0 = np.zeros((2, tp_im.shape[0]))
    for itt2 in range(tp_im.shape[0]):
        # Find indices of non-zero elements
        non_zero_indices = np.where(tp_im[itt2, :] != 0)[0]
        
        # If there are non-zero elements, get first and last
        if non_zero_indices.size > 0:
            first_non_zero = non_zero_indices[0]
            last_non_zero = non_zero_indices[-1]
        else:
            first_non_zero = -1  # Or any placeholder value for rows without non-zero elements
            last_non_zero = -1   # Or any placeholder value for rows without non-zero elements
        
        cell_exists0[:, itt2] = [first_non_zero, last_non_zero]

    sortOrder = sorted(range(cell_exists0.shape[1]), key=lambda i: cell_exists0[0, i])
    ########
        

    # Reorder the array based on the sorted indices
    cell_exists = cell_exists0[:, sortOrder]
    art_cell_exists = cell_exists


    # Re-label
    Mask6 = np.zeros_like(Mask5)
        
    for itt3 in range(len(sortOrder)):
        pix3 = np.where(Mask5 == sortOrder[itt3] + 1)  # here
        Mask6[pix3] = itt3 + 1# reassign

    """
    # Get cell presence 4
    """
    Mask7 = Mask6.copy()
    numbM = im_no
    obj = np.unique(Mask6)
    no_obj1 = int(obj.max())
    A = 1

    tic = time.time()
    tp_im = np.zeros((no_obj1, im_no))
    for cel in range(1, no_obj1 + 1):
        tp_im[cel - 1, :] = ((Mask7 == cel).sum(axis=(0, 1)) != 0).astype(int)
    toc = time.time()
    logger.info(f'GENERAL TRACKING: Elapsed time is {toc - tic} seconds.')

    # plt.figure()
    # plt.imshow(tp_im, aspect='auto', interpolation="nearest")
    # plt.title("Cell Presence Over Time")
    # plt.xlabel("Time")
    # plt.ylabel("Cells")
    # plt.show()

    # Calculate size
    obj = np.unique(Mask7)
    no_obj = int(np.max(obj))
    im_no = Mask7.shape[2]
    all_ob = np.zeros((no_obj, im_no))

    # plt.imshow(Mask7[:,:,im_no-1])
    # MASK7 CONTAINS THE TRACKED SPO-CORRECTED PROSEG MASKS, INCLUDING THE MATING EVENTS 


    tic = time.time()
    for ccell in range(1, no_obj + 1):
        Maa = (Mask7 == ccell)

        for i in range(im_no):
            pix = np.sum(Maa[:, :, i])
            all_ob[ccell-1, i] = pix
    toc = time.time()
    logger.info(f'GENERAL TRACKING: Elapsed time is {toc - tic} seconds.')


    ## COMES FROM: THESE VARIABLES WILL BE USED IN THE LATER STEPS ALONG WITH MASK7
    Art_obj = np.unique(Mask7)
    Art_no_obj = int(np.max(obj))
    Art_im_no = Mask7.shape[2]
    Art_all_ob = all_ob
    
    if transpose_out:
        return np.transpose(Mask7, (2, 0, 1)).astype(np.uint16)  
    return Mask7, art_cell_exists


from typing import Optional, List, Tuple, Union
import torch
import time
import numpy as np
import logging
from .utils import _get_budSeg, _get_matSeg, _get_proSeg, _get_spoSeg, extend_seg_output, fill_empty_arrays, resize_image, save_images_to_directory, synchronize_indices
from .mating import track_mating
from .tetrads import track_tetrads
from .full_lifecycle_utils import track_general_masks
from .correction import correct_mating, correct_proSeg_with_mating, correct_proSeg_with_tetrads
from yeastvision.ims.interpolate import interpolate_intervals, get_interp_labels, deinterpolate
from yeastvision.track.track import track_proliferating
from yeastvision.track.lineage import LineageConstruction
from .FL import FL
from skimage.io import imsave

logger = logging.getLogger(__name__)

def fiest_basic(ims:np.ndarray, interp_intervals:Optional[List[dict]], proSeg_params:Optional[dict], proSeg_weights:Optional[str]=None)->Tuple[np.ndarray[np.uint16], np.ndarray[np.uint8], np.ndarray[np.uint8]]:
    '''
    Implementation 1 of Frame Interpolation Enhanced Single-cell Tracking (FIEST) from the paper 'Deep learning-driven 
    imaging of cell division and cell growth across an entire eukaryotic life cycle' for tracking of timeseries movies of
    yeasts that proliferate asexually. This novel algorithm is based on enhancing the overlap of single 
    cell masks in consecutive images through deep learning video frame interpolation. 
    
    Steps: (1) interpolate with RIFE, (2) segment with proSeg, (3) track, (4) deinterpolate
    
    Args:
        ims (np.ndarray): an (n x h x w) array with n images of h x w dimensions to be interpolated and segmented
        interp_intervals (optional, lists of dict): optional list specifying interpolation levels and intervals, each entry
                                                    of the list should be a dictionary with the keys (start, stop, interp)
                                                    where interp is an integer value in [1,2,3,4]. If none, no interpolation is
                                                    performed
        proSeg_params (optional dict): a dictionary specifying the parameters for segmentation with proSeg, should have the keys 
                                    'mean_diameter,' 'flow_threshold,' and 'cell_probability_threshold.' If none, defaults are used
        proSeg_weights (optional, path): path to weights to use. If none, loads default proSeg weights from the miranda lab. Assumes weights
                                        have been downloaded
                                    
    Returns:
        tuple:
            np.ndarray[uint16] - the tracked masks
            np.ndarray[uint8] - the cell probability masks
            np.ndarray[uint8] - the cell flows masks
    '''
    logger.info("---FIEST Tracking Basic---")
    to_segment = ims
    if interp_intervals:
        logger.info(f"Performing interpolation over intervals:\n {interp_intervals}")
        to_segment, new_intervals = interpolate_intervals(ims, interp_intervals)
        interp_locs = get_interp_labels(len(ims), interp_intervals)

    proSeg, proSeg_params, proSeg_weights = _get_proSeg(proSeg_params, proSeg_weights)
    
    masks, probs, flows = proSeg.run(to_segment, proSeg_params, proSeg_weights)
    masks = track_general_masks(masks, transpose_out=True)
    
    if interp_intervals:
        masks, probs, flows = [deinterpolate(mask_type, interp_locs) for mask_type in [masks, probs, flows]]
    
    torch.cuda.empty_cache()
    del proSeg
    
    return masks, probs, flows
    

def fiest_basic_with_lineage(ims:np.ndarray, interp_intervals:Optional[List[dict]], proSeg_params:Optional[dict], proSeg_weights:Optional[str]=None,
                            budSeg_params:Optional[dict]=None, budSeg_weights:Optional[str]=None)->dict:
    '''
    Implementation 2 of Frame Interpolation Enhanced Single-cell Tracking (FIEST) and lineage reconstruction alogirthm 
    from the paper 'Deep learning-driven imaging of cell division and cell growth across an entire eukaryotic life cycle' 
    for tracking cells and lineages of yeasts that proliferate asexually. This novel algorithm is based on enhancing the overlap of single 
    cell masks and mother-daughter pairs in consecutive images through deep learning video frame interpolation.
    
    Steps: (1) interpolate with RIFE, (2) segment with proSeg, (3) track, (4) construct lineages, (4) deinterpolate
    
    Args:
        ims (np.ndarray): an (n x h x w) array with n images of h x w dimensions to be interpolated and segmented
        interp_intervals (list, optional): optional list specifying interpolation levels and intervals, each entry
                                                    of the list should be a dictionary with the keys (start, stop, interp)
                                                    where interp is an integer value in [1,2,3,4]. If none, no interpolation is
                                                    performed
        proSeg_params (dict, optional): a dictionary specifying the parameters for segmentation with proSeg, should have the keys 
                                    'mean_diameter,' 'flow_threshold,' and 'cell_probability_threshold.' If none, defaults are used
        proSeg_weights (str, optional): path to weights to use. If none, loads default proSeg weights from the miranda lab. Assumes weights
                                        have been downloaded
        budSeg_params (dict, optional): budSeg parameters for budding cell segmentation. See proSeg_params
        budSeg_weights (str, optional): path to weights for budSeg predictions, see proSeg_weights. 
                                    
    Returns:
        dict: a dictionary with the segmentation outputs and lineage data, containing the following keys:
            'cells' (tuple): the output of the cytoplasm (proSeg) model with the following items
                np.ndarray[uint16] - the tracked masks
                np.ndarray[uint8] - the cell probability masks
                np.ndarray[uint8] - the cell flows masks
            'buds' (tuple): the output of the bud segmentation used for lineage constructions:
                np.ndarray[uint16] - the labeled buds
                np.ndarray[uint8] - the bud probability masks
                np.ndarray[uint8] - the bud flows masks
            'lineages' (tuple): the lineage information:
                np.ndarray[uint16] - cell daughters array
                pd.DataFrame - a DataFrame containing cell life data with columns 'birth', 'death', 'mother', and 'confidence'.
    '''
    logger.info("---FIEST Tracking With Lineage Reconstruction---")
    to_segment = ims
    if interp_intervals:
        logger.info(f"Performing interpolation over intervals:\n {interp_intervals}")
        to_segment, new_intervals = interpolate_intervals(ims, interp_intervals)
        interp_locs = get_interp_labels(len(ims), interp_intervals)

    proSeg, proSeg_params, proSeg_weights = _get_proSeg(proSeg_params, proSeg_weights)
    budSeg, budSeg_params, budSeg_weights = _get_budSeg(budSeg_params, budSeg_weights)
    
    masks, probs, flows = proSeg.run(to_segment, proSeg_params, proSeg_weights)
    masks = track_general_masks(masks, transpose_out=True)

    
    buds, bud_probs, bud_flows = budSeg.run(to_segment, budSeg_params, budSeg_weights)
    
    lineage = LineageConstruction(masks, buds, forwardskip=3)
    daughters, mothers = lineage.computeLineages()
    
    if interp_intervals:
        masks, probs, flows = [deinterpolate(mask_type, interp_locs) for mask_type in [masks, probs, flows]]
        buds, bud_probs, bud_flows = [deinterpolate(mask_type, interp_locs) for mask_type in [buds, bud_probs, bud_flows]]
    
    torch.cuda.empty_cache()
    del proSeg
    del budSeg
    
    return {"cells": (masks, probs, flows), 
            "buds": (buds, bud_probs, bud_flows),
            "lineages": (daughters, mothers)
    }


def fiest_full_lifecycle(ims: np.ndarray, interp_intervals:Optional[List[dict]]=None, proSeg_params:Optional[dict]=None, proSeg_weights:Optional[str]=None, 
                         matSeg_params:Optional[dict]=None, matSeg_weights:Optional[str]=None, spoSeg_params: Optional[dict]=None, spoSeg_weights:Optional[str]=None,
                         mat_start:Optional[int]=0, mat_stop:Optional[int]=None, spo_start:Optional[int]=0, spo_stop:Optional[int]=None, shock_period:Optional[List[int]]=None):
    '''
    Implementation of Frame Interpolation Enhanced Single-cell Tracking (FIEST) from the paper 'Deep learning-driven 
    imaging of cell division and cell growth across an entire eukaryotic life cycle' for tracking of timeseries movies of
    yeasts that mate, sporulate, and proliferate asexually. This novel algorithm is based on enhancing the overlap of single 
    cell masks in consecutive images through deep learning video frame interpolation. 
    
    Steps: 
    (1) segment ims with matSeg, proSeg, and spoSeg
    (2) track using a novel full-lifecycle track method based frame overlap and corrections that leverage all three outputs
    (3) de-interpolate all outputs
    
    Args:
        ims (np.ndarray): an (n x h x w) array with n images of h x w dimensions to be interpolated and segmented
        interp_intervals (list, optional): optional list specifying interpolation levels and intervals, each entry
                                                    of the list should be a dictionary with the keys (start, stop, interp)
                                                    where interp is an integer value in [1,2,3,4]. If none, no interpolation is
                                                    performed
        proSeg_params (dict, optional): a dictionary specifying the parameters for segmentation with proSeg, should have the keys 
                                    'mean_diameter,' 'flow_threshold,' and 'cell_probability_threshold.' If none, defaults are used
        proSeg_weights (str, optional): path to weights to use. If none, loads default proSeg weights from the Miranda lab. Assumes weights
                                        have been downloaded
        matSeg_params (dict, optional): matSeg parameters for mating cell segmentation. See proSeg_params
        matSeg_weights (str, optional): path to weights for matSeg predictions, see proSeg_weights.  
        spoSeg_params (dict, optional): spoSeg parameters for tetrad segmentation. See proSeg_params
        spoSeg_weights (str, optional): path to weights for spoSeg predictions, see proSeg_weights.  
        mat_start (int, optional): the first index (0-indexed) of mating cells. Default is 0
        mat_stop (int, optional): the last index (exclusive, 0-indexed) of mating cells in ims. If none, then mat_stop defaults to len(ims).
        spo_start (int, optional): the first index (0-indexed) of sporulating cells (tetrads) in ims. Default is 0.
        spo_stop (int, optional): the last index (exclusive, 0-indexed) of sporulating cells (tetrads) in ims. If none, defaults to len(ims).
        shock_period (list of integers, optional): a two number list specifying to python indeces (0-indexed, exclusive) of the shock_period in the experiments.
                                                    Leave as none if no shock_period applies to the movie.
        
        
                                    
    Returns:
        dict: a dictionary with tracked outputs, probability, and flow images. Please noe that probability and flow images have not been corrected.  
        This dictionary contains the following keys:
            'cells' (tuple): output of proSeg segmentation and tracking
                np.ndarray[uint16] - the tracked general (proSeg) masks with mating events removed
                np.ndarray[uint8] - the raw proSeg cell probability masks
                np.ndarray[uint8] - the raw proSeg cell flows masks 
            'mating' (tuple): output of matSeg segmentation and tracking
                np.ndarray[uint16] - the mating tracked masks after correction
                np.ndarray[uint8] - the raw matSeg mating probability masks
                np.ndarray[uint8] - the raw matSeg mating flows masks
            'spores' (tuple): output of spoSeg segmentation and tracking:
                np.ndarray[uint16] - the tracked tetrad masks after correction
                np.ndarray[uint8] - the raw spoSeg tetrad probability masks
                np.ndarray[uint8] - the raw spoSeg tetrad flows masks
    '''

    logger.info("---FIEST Full Lifecycle Tracking---")
    to_segment = ims
    if interp_intervals:
        logger.info(f"Performing interpolation over intervals:\n {interp_intervals}")
        to_segment, new_intervals = interpolate_intervals(ims, interp_intervals)
        interp_locs = get_interp_labels(len(ims), interp_intervals)

    
    #logger.info("Fetching models and weight files")
    proSeg, proSeg_params, proSeg_weights = _get_proSeg(proSeg_params, proSeg_weights)
    spoSeg, spoSeg_params, spoSeg_weights = _get_spoSeg(spoSeg_params, spoSeg_weights)
    matSeg, matSeg_params, matSeg_weights = _get_matSeg(matSeg_params, matSeg_weights)

    #logger.info("")
    masks, probs, flows = proSeg.run(to_segment, proSeg_params, proSeg_weights)

    if mat_stop is None:
        mat_stop = len(ims)
    if spo_stop is None:
        spo_stop = len(ims)
    
    #logger.info(f"Segment with matseg from")
    mat_out = matSeg.run(to_segment[mat_start:mat_stop], matSeg_params, matSeg_weights)
    mat_out_correct_len = extend_seg_output(to_segment, mat_out, mat_start, mat_stop)
    
    #logger.info()
    spo_out = spoSeg.run(to_segment[spo_start:spo_stop], spoSeg_params, spoSeg_weights)
    spo_out_correct_len = extend_seg_output(to_segment, spo_out, spo_start, spo_stop)

    proSeg_tracked, matSeg_tracked, spoSeg_tracked = track_full_lifecycle(masks, mat_out_correct_len[0], spo_out_correct_len[0],
                                                                          [spo_start, spo_stop], [mat_start, mat_stop], len(to_segment), 
                                                                          shock_period=shock_period)

    if interp_intervals:
        proSeg_tracked, probs, flows = [deinterpolate(mask_type, interp_locs) for mask_type in [proSeg_tracked, probs, flows]]
        matSeg_tracked, mat_probs, mat_flows = [deinterpolate(mask_type, interp_locs) for mask_type in [matSeg_tracked, mat_out_correct_len[1], mat_out_correct_len[2]]]
        spoSeg_tracked, spo_probs, spo_flows = [deinterpolate(mask_type, interp_locs) for mask_type in [spoSeg_tracked, spo_out_correct_len[1], spo_out_correct_len[2]]]
        
    else:
        mat_probs, mat_flows = mat_out_correct_len[0], mat_out_correct_len[1]
        spo_probs, spo_flows = spo_out_correct_len[0], spo_out_correct_len[1]
        

    torch.cuda.empty_cache()
    del proSeg
    del spoSeg
    del matSeg

    return {"cells": (proSeg_tracked, probs, flows),
            "mating": (matSeg_tracked, mat_probs, mat_flows),
            "spores": (spoSeg_tracked, spo_probs, spo_flows)
            }
    
def track_full_lifecycle(proSeg:Union[np.ndarray, List[np.ndarray]], mating: Union[np.ndarray, List[np.ndarray]], tetrads: Union[np.ndarray, List[np.ndarray]], 
                         tetrad_interval:List[int], mating_interval: List[int], movie_length:int, shock_period:Optional[List[int]]=None):
    '''
    Performs the full-lifecycle tracking of yeast cells using the general (proSeg) detections, mating (matSeg) detections, and tetrad (spoSeg) detections using a novel algorithm developed at the Miranda Lab
    at NCSU. The algorithm is as follows:
    
    1) track tetrads
    2) correct general masks with the tracked tetrads
    3) track the corrected general masks
    4) track mating events
    5) correct the general masks using the tracked mating events

    Args:
        proSeg (np.ndarray or list of np.ndarrays) - the general masks either as 3D array (time x height x width) or a list of numpy arrays (height x width)
        proSeg (np.ndarray or list of np.ndarrays) - the general masks either as 3D array (time x height x width) or a list of numpy arrays (height x width)
        proSeg (np.ndarray or list of np.ndarrays) - the general masks either as 3D array (time x height x width) or a list of numpy arrays (height x width)
    Returns
        tuple: a tuple of three 3D numpy arrays of the same shape. In the case that the tetrad/mating interval did not encompass the whole movie,
        the tetrad/mating masks will be padded with 0 arrays to reach the same length as the full-length movie. All arrays in the tuple will have the same shape (timepoints x rows x cols). The ordering of the tuple is as follows:
            np.ndarray[uint16] - the tracked general masks with spores correct
            np.ndarray[uint16] - tracked mating cells
            np.ndarray[uint16] - tracked sporulating cells (tetrads)
        '''
    if shock_period is None:
        shock_period=[0,0]
    
    logger.info("FULL LIFECYCLE TRACKING: 6 Steps")
    logger.info("STEP 1/6: Track Tetrads")
    TET_obj, TET_exists, TETmasks = track_tetrads(tetrads, tetrad_interval, movie_length, shock_period) # step2
    
    logger.info("STEP 2/6: Track Mating Cells")
    Matmasks, mat_no_obj, Mat_cell_data, cell_data = track_mating(mating, mating_interval, shock_period) # step3
    #save_images_to_directory(Matmasks, "/home/berk/code/yeastvision/test/step3")
    
    
    logger.info("STEP 3: Correct General Masks")
    proSeg_corrected_tetrads = correct_proSeg_with_tetrads(proSeg, shock_period, TET_obj, TET_exists, TETmasks) # step4
    
    logger.info("STEP 4/6: Track General Masks ")
    Mask7, art_cell_exists = track_general_masks(proSeg_corrected_tetrads) # step 5
    
    logger.info("STEP 5/6: Correct Mating Cells")
    Matmasks, final_mat_cell_data, mat_no_obj = correct_mating(Matmasks, Mask7, mat_no_obj, Mat_cell_data, cell_data) # step6
    #save_images_to_directory(Matmasks, "/home/berk/code/yeastvision/test/step6")
    
    logger.info("STEP 6/6: Correct General Masks again")
    final_art_masks = correct_proSeg_with_mating(Matmasks, Mask7, art_cell_exists, mat_no_obj, final_mat_cell_data) # step7
    
    logger.info("Post-Processing Outputs...padding, resizing, transposing")
    mating_tracks = np.array(fill_empty_arrays(Matmasks), dtype=np.uint16)
    
    tetrad_resized = [resize_image(image, final_art_masks[:,:,0].shape) for image in fill_empty_arrays(TETmasks)]
    tetrad_tracks = np.zeros_like(mating_tracks)
    tetrad_tracks[: tetrad_interval[1]] = np.array(tetrad_resized) 
    final_art_masks = np.transpose(final_art_masks, (2, 0, 1)).astype(np.uint16)  
    
    logger.info("Post-Processing Tetrad Masks...Synchronizing tetrad cell values with the general track") 
    tetrad_tracks = synchronize_indices(final_art_masks, tetrad_tracks)

    return final_art_masks, mating_tracks, tetrad_tracks



# This section utilizes the RIFE (Real-Time Intermediate Flow Estimation) method
# described in:
# Huang, Z., Zhang, T., Heng, W., Shi, B., & Zhou, S. (2022). Real-time intermediate 
# flow estimation for video frame interpolation. In Proceedings of the European 
# Conference on Computer Vision (ECCV).
# The original code can be found at [https://github.com/megvii-research/ECCV2022-RIFE


from PyQt5 import QtGui, QtCore
from PyQt5.QtWidgets import (QTreeWidgetItem, QPushButton, QDialog,
                        QDialogButtonBox, QLineEdit, QFormLayout, QCheckBox,  QSpinBox, QDoubleSpinBox, QLabel, 
                            QWidget, QComboBox, QGridLayout, QHBoxLayout, QHeaderView, QVBoxLayout, QMessageBox,
                            QTreeWidget, QButtonGroup, QFrame, QRadioButton)
from PyQt5.QtCore import QTimer, Qt, pyqtSignal
import numpy as np
import os
from yeastvision.parts.guiparts import *
from yeastvision.plot.plot import PlotProperty
import math
from yeastvision.data.ims import InterpolatedChannel

class ChoiceDialog(QDialog):
    def __init__(self, choices, title, parent=None):
        super().__init__(parent)
        
        self.setWindowTitle(title)
        
        self.layout = QVBoxLayout()
        self.button_group = QButtonGroup(self)
        
        # Create radio buttons for each choice
        for choice in choices:
            radio_button = QRadioButton(choice)
            self.layout.addWidget(radio_button)
            self.button_group.addButton(radio_button)
        
        # Add OK and Cancel buttons
        self.ok_button = QPushButton("OK")
        self.cancel_button = QPushButton("Cancel")
        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        
        self.layout.addWidget(self.ok_button)
        self.layout.addWidget(self.cancel_button)
        
        self.setLayout(self.layout)
        
    def get_choice(self):
        # Returns the selected choice, or None if no selection is made
        selected_button = self.button_group.checkedButton()
        if selected_button:
            return selected_button.text()
        return None

class ComboBoxDialog(QDialog):
    def __init__(self, choices, label, title, parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.setModal(True)
        self.setLayout(QVBoxLayout())

        self.label = QLabel(label)
        self.layout().addWidget(self.label)

        self.comboBox = QComboBox()
        self.comboBox.addItems(choices)
        self.layout().addWidget(self.comboBox)

        self.okButton = QPushButton("OK")
        self.okButton.clicked.connect(self.accept)
        self.layout().addWidget(self.okButton)

    def getSelection(self):
        return self.comboBox.currentText()

class SimpleTextDialog(QDialog):
    def __init__(self, text, parent=None):
        super().__init__(parent)
        
        self.setWindowTitle("Information")
        self.setGeometry(100, 100, 400, 200)

        layout = QVBoxLayout()

        # Text label
        text_label = QLabel(text)
        layout.addWidget(text_label)

        # Ok and Cancel buttons
        button_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK")
        self.ok_button.clicked.connect(self.accept)
        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)

        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)

        layout.addLayout(button_layout)

        self.setLayout(layout)

class IntervalSelectionDialog(QDialog):
    def __init__(self, intervals, maxT, windowTitle, parent=None, labels = None, presetT1 = 0, presetT2 = None):
        super().__init__(parent)
        self.intervals = intervals
        self.maxT = maxT

        self.setWindowTitle(windowTitle)
        #self.setGeometry(100, 100, 400, 200)

        layout = QVBoxLayout()

        if labels:
            if isinstance(labels, list):
                for label in labels:
                    layout.addWidget(QLabel(label))
            else:
                layout.addWidget(QLabel(labels))

        # Spin boxes for start and end indices
        spin_layout = QHBoxLayout()
        self.start_spinbox = QSpinBox()
        self.start_spinbox.setRange(0, maxT)
        self.start_spinbox.setValue(presetT1)
        self.start_spinbox.valueChanged.connect(self._update_end_spinbox_range)

        self.end_spinbox = QSpinBox()
        self.end_spinbox.setRange(0, maxT + 1)
        if presetT2 is None:
            presetT2 = presetT1
        self.end_spinbox.setValue(presetT2)
        self.start_spinbox.valueChanged.connect(self._sync_end_spinbox)

        spin_layout.addWidget(QLabel("Start:"))
        spin_layout.addWidget(self.start_spinbox)
        spin_layout.addWidget(QLabel("End:"))
        spin_layout.addWidget(self.end_spinbox)

        layout.addLayout(spin_layout)

        # Ok and Cancel buttons
        button_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK")
        self.ok_button.clicked.connect(self.accept)
        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)

        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)

        layout.addLayout(button_layout)

        self.setLayout(layout)

    def _update_end_spinbox_range(self):
        self.end_spinbox.setMinimum(self.start_spinbox.value() + 1)

    def _sync_end_spinbox(self):
        if self.end_spinbox.value() < self.start_spinbox.value():
            self.end_spinbox.setValue(self.start_spinbox.value() + 1)

    def get_selected_interval(self):
        return self.start_spinbox.value(), self.end_spinbox.value()

class InterpolationIntervalWidget(QFrame):
    removed = pyqtSignal(object)
    
    def __init__(self, parent=None, max_frame=10000):
        super().__init__(parent)
        
        self.start_frame_spinbox = QSpinBox()
        self.start_frame_spinbox.setRange(0, max_frame - 1)  # Set realistic frame range
        
        self.end_frame_spinbox = QSpinBox()
        self.end_frame_spinbox.setRange(1, max_frame)  # Set realistic frame range
        
        self.end_frame_spinbox.valueChanged.connect(self.update_start_frame_max)
        
        self.interpolation_combo = QComboBox()
        self.interpolation_combo.addItems(["2x", "4x", "8x", "16x"])
        
        self.remove_button = QPushButton("Remove")
        self.remove_button.clicked.connect(self.remove_interval)
        
        layout = QHBoxLayout()
        layout.addWidget(QLabel("Start Frame:"))
        layout.addWidget(self.start_frame_spinbox)
        layout.addWidget(QLabel("End Frame:"))
        layout.addWidget(self.end_frame_spinbox)
        layout.addWidget(QLabel("Interpolation:"))
        layout.addWidget(self.interpolation_combo)
        layout.addWidget(self.remove_button)
        
        self.setLayout(layout)
        
    def remove_interval(self):
        self.setParent(None)
        self.removed.emit(self)  # Emit the removed signal

    def set_max_frame(self, max_frame):
        self.start_frame_spinbox.setMaximum(max_frame - 1)
        self.end_frame_spinbox.setMaximum(max_frame)
        self.update_start_frame_max()
        if self.end_frame_spinbox.value() > max_frame:
            self.end_frame_spinbox.setValue(max_frame)

    def update_start_frame_max(self):
        self.start_frame_spinbox.setMaximum(self.end_frame_spinbox.value() - 1)
    
    def get_interval_data(self):
        return {
            "start": self.start_frame_spinbox.value(),
            "stop": self.end_frame_spinbox.value(),
            "interp": int(math.log(int(self.interpolation_combo.currentText().replace("x", "")), 2))
        }

class InterpolationDialog(QDialog):
    def __init__(self, text = None, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Select Interpolation Intervals")
        
        self.layout = QVBoxLayout()

        if text:
            self.layout.addWidget(text)
        
        self.channel_combo = self.add_parent_combobox(self.parent.channelSelect)
        self.channel_combo.currentIndexChanged.connect(self.channel_select_change)
        
        self.layout.addWidget(QLabel("Channel to Interpolate:"))
        self.layout.addWidget(self.channel_combo)

        self.error_label = QLabel("")
        self.error_label.setStyleSheet("color: red")
        self.layout.addWidget(self.error_label)
        
        self.add_interval_button = QPushButton("Add Interval")
        self.add_interval_button.clicked.connect(self.add_interval)
        
        self.dialog_buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            parent=self
        )
        self.dialog_buttons.accepted.connect(self.check_intervals)
        self.dialog_buttons.rejected.connect(self.reject)
        
        self.layout.addWidget(self.add_interval_button)
        self.layout.addWidget(self.dialog_buttons)
        
        self.setLayout(self.layout)
        
        self.intervals = []
        self.channel_select_change()
    
    def remove_all_intervals(self):
        for interval_widget in self.intervals:
            interval_widget.setParent(None)
        self.intervals.clear()

    def disable_and_clear_interval_add(self):
        self.remove_all_intervals()
        self.add_interval_button.setDisabled(True)
    
    def enable_interval_add(self):
        self.add_interval_button.setDisabled(False)

    def add_interval(self):
        max_frame = self.get_max_frame()
        interval_widget = InterpolationIntervalWidget(max_frame=max_frame)
        interval_widget.removed.connect(self.remove_interval)  # Connect the signal
        self.intervals.append(interval_widget)
        self.layout.insertWidget(self.layout.count() - 3, interval_widget)

    def remove_interval(self, interval_widget):
        if interval_widget in self.intervals:
            self.intervals.remove(interval_widget)
            interval_widget.deleteLater()

    def channel_select_change(self):
        self.channel_obj = self.parent.get_channel_obj_from_name(self.channel_combo.currentText())
        if isinstance(self.channel_obj, InterpolatedChannel):
            self.disable_and_clear_interval_add()
            self.showMessage(f"{self.channel_obj.name} has already been interpolated")
        else:
            self.clearMessages()
            self.enable_interval_add()
            self.update_max_frame()

    def update_max_frame(self):
        max_frame = self.get_max_frame()
        if max_frame == 0:
            self.disable_and_clear_interval_add()
            self.showMessage("Current Channel has only one image.")
        else:
            self.add_interval_button.setDisabled(False)
            self.clearMessages()
            for interval_widget in self.intervals:
                interval_widget.set_max_frame(max_frame)

    def add_parent_combobox(self, parentComboBox):
        channelSelect = QComboBox()
        channelSelect.setFixedWidth(200)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        return channelSelect

    def get_max_frame(self):
        return self.channel_obj.max_t()

    def check_intervals(self):
        new_intervals = []
        for interval_widget in self.intervals:
            start_frame = interval_widget.start_frame_spinbox.value()
            end_frame = interval_widget.end_frame_spinbox.value()
            interp = interval_widget.interpolation_combo.currentText()
            max_frame = self.get_max_frame()

            if start_frame < 0 or start_frame >= max_frame:
                self.showMessage(f"Start frame {start_frame} is out of range.")
                return
            if end_frame <= start_frame or end_frame > max_frame:
                self.showMessage(f"End frame {end_frame} is out of range or not greater than start frame.")
                return

            new_intervals.append((start_frame, end_frame, interp))

        # Check for overlapping intervals
        new_intervals.sort()  # Sort intervals by start frame
        for i in range(len(new_intervals) - 1):
            if new_intervals[i][1] > new_intervals[i + 1][0]:
                self.showMessage("Intervals overlap.")
                return

        self.accept()
        self.clearMessages()
    
    def get_data(self):
        channel = self.channel_combo.currentText()
        intervals = [interval_widget.get_interval_data() for interval_widget in self.intervals]
        intervals = sorted(intervals, key=lambda x: x['start'])
        return channel, intervals

    def showMessage(self, message):
        self.error_label.setText(message)

    def clearMessages(self):
        self.showMessage("") 

class TimedPopup(QDialog):
    def __init__(self, text, seconds, parent=None):
        super().__init__(parent)
        self.setWindowTitle('Popup')
        label = QLabel(text)
        layout = QVBoxLayout()
        layout.addWidget(label)
        self.setLayout(layout)
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.accept)
        self.timer.start(seconds*100)  # 10 seconds timeout

    def closeEvent(self, event):
        self.timer.stop()
        if self.parent():
            self.parent().popup = None
        super().closeEvent(event)

class FigureDialog(QDialog):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.layout = QGridLayout()
        self.row = 0

        self.label_widgets = {}
        self.label_names = self.parent.experiment().get_label_names()
        self.make_top_row()
        for i, label_name in enumerate(self.label_names):
            self.add_row(label_name, i)

    
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.layout.addWidget(self.buttonBox, self.row, 0)
        
        self.setLayout(self.layout)

    def make_top_row(self):
        for i, header in enumerate(["label name", "contour", "full label", "order"]):
            self.layout.addWidget(QLabel(header), self.row, i,1,1)
        self.row+=1
    
    def add_row(self, label_name, i):
        self.label_widgets[label_name] = {}
        self.layout.addWidget(QLabel(label_name), self.row, 0,1,1)

        contour_check = QCheckBox()
        self.label_widgets[label_name]["contours"] = contour_check
        self.layout.addWidget(contour_check, self.row,1,1,1)

        label_check = QCheckBox()
        self.label_widgets[label_name]["labels"] = label_check
        self.layout.addWidget(label_check, self.row,2,1,1)

        order_spin = QSpinBox()
        order_spin.setMaximum(len(self.label_names)-1)
        order_spin.setValue(i)
        self.label_widgets[label_name]["order"] = order_spin
        self.layout.addWidget(order_spin, self.row, 3,1,1)

        self.row+=1
    
    def get_data(self):
        data = self.label_widgets.copy()
        for name in data.keys():
            data[name]["contours"] = data[name]["contours"].isChecked()
            data[name]["labels"] = data[name]["labels"].isChecked()
            data[name]["order"] = data[name]["order"].value()
        return data


class DirReaderDialog(QDialog):
    def __init__(self, dir, fnames):
        super().__init__()

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.formLayout = QFormLayout()
        self.formLayout.addRow(QLabel(f"Load Custom Channels\n{dir}"))
        self.setLayout(self.formLayout)


        # Create the QSpinBox
        self.spinBox = QSpinBox(self)
        self.spinBox.setMinimumWidth(100)
        self.spinBox.setMinimum(0) # Set minimum value
        self.spinBox.valueChanged.connect(self.update_line_edits) # Connect signal to slot
        self.formLayout.addRow("number of channels:", self.spinBox)
        self.lineEditLayout = QFormLayout()
        self.formLayout.addRow(self.lineEditLayout)
        self.formLayout.addWidget(self.buttonBox)

        # Initialize list to hold QLineEdits
        self.line_edits = []
    

    def update_line_edits(self, value):
        # Check if the value has increased or decreased
        if value > len(self.line_edits):
            # Value increased, create new QLineEdit(s)
            for i in range(len(self.line_edits), value):
                line_edit = QLineEdit(self)
                line_edit.setToolTip("a sequence of characters only found in this channels filename")
                self.line_edits.append(line_edit)
                self.lineEditLayout.addRow(f"channel {i} identifier", line_edit)
        elif value < len(self.line_edits):
            # Value decreased, remove QLineEdit(s)
            for i in range(len(self.line_edits)-1, value-1, -1):
                self.lineEditLayout.removeRow(self.line_edits[i]) # Remove QLineEdit from layout
                self.line_edits[i].deleteLater() # Delete QLineEdit from memory
                self.line_edits.pop(i) # Remove QLineEdit from list
    
    def getData(self):
        return [line_edit.text() for line_edit in self.line_edits]
    
class GeneralParamDialog(QDialog):
    def __init__(self, hyperparamDict, paramtypes, windowName, parent, labelSelects = None, channelSelects = None):
        
        super().__init__(parent)
        self.parent = parent
        self.params = hyperparamDict
        self.types = paramtypes

        self.dropDowns = {}
        self.checkBoxes = {}
        self.lineEdits = {}
        self.spinBoxes = {}
        self.dSpinBoxes = {}

        if self.params:
            self.splitHyperParams()

        self.labelSelects = labelSelects
        self.channelSelects = channelSelects

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.message = QLabel(f"Enter parameters for {windowName}")

        self.formLayout = QFormLayout()
        self.formLayout.addRow(self.message)
        self.populateFormLayout()
        self.formLayout.addWidget(self.buttonBox)
        self.setLayout(self.formLayout)
        self.errorAlrShown = False
    
    def splitHyperParams(self):
        for paramName, paramType in zip(list(self.params.keys()),self.types) :
            if paramType == "drop":
                self.dropDowns[paramName] = ""
            elif paramType is bool:
                self.checkBoxes[paramName] = self.params[paramName]
            elif paramType is None:
                self.lineEdits[paramName] = self.params[paramName]
            elif paramType is int:
                self.spinBoxes[paramName] = self.params[paramName]
            elif paramType is float:
                self.dSpinBoxes[paramName] = self.params[paramName]
    
    def populateFormLayout(self):
        self.dropDownData  = {}
        if self.channelSelects:
            for channelSelectName in self.channelSelects:
                self.dropDownData[channelSelectName] = self.addParentComboBox(self.parent.channelSelect, name = channelSelectName)
        if self.labelSelects:
            for labelSelectName in self.labelSelects:
                self.dropDownData[labelSelectName] = self.addParentComboBox(self.parent.labelSelect, name = labelSelectName)
        
        # check boxes
        hLayout = QHBoxLayout()
        self.checkBoxData = {}
        if self.checkBoxes:
            for checkName, defaultVal in self.checkBoxes.items():
                checkBox = QPushButton(checkName)
                checkBox.setCheckable(True)
                hLayout.addWidget(checkBox)
                checkBox.setChecked(defaultVal)
                self.checkBoxData[checkName] = checkBox
            self.formLayout.addRow(hLayout)

        formDict = self.lineEdits
        self.lineEditData = {}
        for labelName,defaultVal in formDict.items():
            lineEdit  = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[labelName] = lineEdit
            self.formLayout.addRow(self.produceLabel(labelName), lineEdit)
        
        self.spinBoxData = {}
        for boxName, defaultVal in self.spinBoxes.items():
            spinBox = QSpinBox(self)
            spinBox.setValue(int(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        for boxName, defaultVal in self.dSpinBoxes.items():
            spinBox = QDoubleSpinBox(self)
            spinBox.setValue(float(defaultVal))
            spinBox.setMinimum(-1.0)
            spinBox.setMaximum(1000.0)
            spinBox.setSingleStep(0.1)
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)

    
    def addParentComboBox(self, parentComboBox, name = "Channel", addNA = False):
        channelSelect = QComboBox()
        channelSelect.setFixedWidth(200)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        self.formLayout.addRow(self.produceLabel(name), channelSelect)
        return channelSelect

    def produceLabel(self, text):
        label = QLabel(text)
        return label
    
    def getData(self):
        dropDownData = {k:v.currentText() for k,v in self.dropDownData.items()}
        lineEditData = {k:float(v.text()) for k,v in self.lineEditData.items()}
        boolData = {k:v.isChecked() for k,v in self.checkBoxData.items()}
        spinData = {k:v.value() for k,v in self.spinBoxData.items()}
        return dropDownData | lineEditData | boolData | spinData

class ModelParamDialog(GeneralParamDialog):
    def __init__(self, hyperparamDict, paramtypes, modelName, parent = None):
        if parent.experiment().has_labels() and modelName != "artilife":
            curr_label = parent.label().name
            hyperparamDict[f"insert into {curr_label}"] = False
            paramtypes.append(bool)
        super().__init__(hyperparamDict, paramtypes, modelName, parent)

    def populateFormLayout(self):
        self.dropDownData  = {}
        self.dropDownData["Channel"] = self.addParentComboBox(self.parent.channelSelect, name = "Channel to Segment")
        # order is important here as self.dropDownData["T Start"]/["T Stop"]
            # must exist before addTValues and indexChange can be called
        self.addTDropDown()
        self.dropDownData["Channel"].currentIndexChanged.connect(self.channelSelectIndexChanged)
        self.channelSelectIndexChanged()
        
        if self.dropDowns:
            for dropDownName in self.dropDowns.keys():
                self.dropDownData[dropDownName] = self.addParentComboBox(self.parent.channelSelect, name = dropDownName)
        
        # check boxes
        hLayout = QHBoxLayout()
        self.checkBoxData = {}
        if self.checkBoxes:
            for checkName, defaultVal in self.checkBoxes.items():
                checkBox = QPushButton(checkName)
                checkBox.setCheckable(True)
                hLayout.addWidget(checkBox)
                checkBox.setChecked(defaultVal)
                self.checkBoxData[checkName] = checkBox
            self.formLayout.addRow(hLayout)

        formDict = self.lineEdits
        self.lineEditData = {}
        for labelName,defaultVal in formDict.items():
            lineEdit  = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[labelName] = lineEdit
            self.formLayout.addRow(self.produceLabel(labelName), lineEdit)
    
        self.spinBoxData = {}
        for boxName, defaultVal in self.spinBoxes.items():
            spinBox = QSpinBox(self)
            spinBox.setValue(int(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        for boxName, defaultVal in self.dSpinBoxes.items():
            spinBox = QDoubleSpinBox(self)
            spinBox.setMinimum(-1.0)
            spinBox.setSingleStep(0.1)
            spinBox.setMaximum(1000.0)
            spinBox.setValue(float(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)

    def channelSelectIndexChanged(self):
        channelIndex = self.parent.channelSelect.findText(self.dropDownData["Channel"].currentText())
        maxT = self.parent.experiment().channels[channelIndex].max_t()
        self.dropDownData["T Start"].clear()
        self.dropDownData["T Stop"].clear()
        self.addTValues(maxT)

    def addTimeDropDown(self, name):
        timeSelect = QComboBox()
        timeSelect.setFixedWidth(120)
        self.formLayout.addRow(self.produceLabel(name), timeSelect)
    
    def addTDropDown(self):
        hbox = QHBoxLayout()
        timeSelectStart, timeSelectStop = QComboBox(), QComboBox()
        timeSelectStart.setFixedWidth(120)
        timeSelectStart.setToolTip("Start time point")
        timeSelectStop.setFixedWidth(120)
        timeSelectStop.setToolTip("Stop Time Point (Inclusive)")
        hbox.addWidget(timeSelectStart)
        hbox.addWidget(timeSelectStop)

        self.dropDownData["T Start"] = timeSelectStart
        self.dropDownData["T Stop"] = timeSelectStop
        self.formLayout.addRow(QLabel("Time Start/Stop"), hbox)
    
    def addTValues(self, maxT, prefix = "T "):
        validTs = list(range(maxT+1))
        validTs = [str(t) for t in validTs]
        self.dropDownData[f"{prefix}Start"].addItems(validTs)
        self.dropDownData[f"{prefix}Start"].setCurrentIndex(self.parent.tIndex)
        self.dropDownData[f"{prefix}Stop"].addItems(validTs)
        self.dropDownData[f"{prefix}Stop"].setCurrentIndex(self.parent.tIndex)
    
    def showErrorMessage(self):
        if not self.errorAlrShown:
            self.message.setText(self.message.text() + "\nEnsure T Start<T Stop and All Line Entries are Numeric")
            self.errorAlrShown = True

    def accept(self):
        if self.validateLineData() and self.validateTData():
            super().accept()
        else:
            self.showErrorMessage()
    
    def validateTData(self):
        data = self.dropDownData
        valid = int(data["T Start"].currentText())<=int(data["T Stop"].currentText())
        return valid

    def validateLineData(self):
        data = {k:v.text() for k,v in self.lineEditData.items()}
        i = 0
        vals = list(data.values())
        valid = True
        while i<len(vals) and valid:
            valid = self.isNumber(vals[i]) and float(vals[i])>=0.0
            i+=1
        return valid
    
    def isNumber(self, num):
        """ Returns True is string is a number. """
        # must remove periods to ensure decimals result in True
        return num.replace(".","").isdigit()


class FlouParamDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.message = QLabel(f"Parameters for blob detection")

        self.formLayout = QFormLayout()
        self.formLayout.addRow(self.message)
        self.populateFormLayout()
        self.formLayout.addWidget(self.buttonBox)
        self.setLayout(self.formLayout)
        self.errorAlrShown = False
    
    def populateFormLayout(self):
        self.checkBoxData = {}
        useExampleCheck = QPushButton("extract params from current mask")
        useExampleCheck.setEnabled(np.any(self.parent.getCurrMaskSet() > 0))
        useExampleCheck.setCheckable(True)
        useExampleCheck.setChecked(False)
        useExampleCheck.clicked.connect(self.exampleCheckClicked)
        self.checkBoxData["extract from curr mask"] = useExampleCheck
        self.formLayout.addWidget(useExampleCheck)

        self.dropDownData = {}
        self.dropDownData["channel"] = self.addParentComboBox(self.parent.channelSelect, name = "Channel to Segment")
        self.dropDownData["label"] = self.addParentComboBox(self.parent.labelSelect, name = "Filter Masks")
        self.dropDownData["label"].addItem("")
        self.dropDownData["label"].setCurrentIndex(-1)
        self.dropDownData["label"].setToolTip("Anything segmented outside 'Filter Mask' will be set to 0")
        contourCheck = QPushButton("use contours")
        contourCheck.setEnabled(np.any(self.parent.getCurrMaskSet() > 0))
        contourCheck.setCheckable(True)
        if contourCheck.isCheckable:
            contourCheck.setChecked(self.parent.contourOn)
        self.formLayout.addWidget(contourCheck)
        self.checkBoxData["use contours"] = contourCheck

        self.addTDropDown()
        self.dropDownData["channel"].currentIndexChanged.connect(self.dIndexChanged)
        self.dropDownData["label"].currentIndexChanged.connect(self.dIndexChanged)
        self.dIndexChanged()

        self.lineEditData = {}
        for lineEditName, defaultVal in zip(["min area", "max area", "threshold"], [20,50,0.08]):
            lineEdit = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[lineEditName] = lineEdit
            self.formLayout.addRow(self.produceLabel(lineEditName), lineEdit)

    def exampleCheckClicked(self):
        enabled = not self.checkBoxData["extract from curr mask"].isChecked()
        for checkBox in self.lineEditData.values():
            checkBox.setEnabled(enabled)

    def dIndexChanged(self):
        channelIndex = self.parent.channelSelect.findText(self.dropDownData["channel"].currentText())
        labelIndex = self.parent.labelSelect.findText(self.dropDownData["label"].currentText())

        if labelIndex == -1:
            maxT = self.parent.imData.maxTs[channelIndex]
        else:
            maxT = min(self.parent.imData.maxTs[channelIndex], self.parent.maskData.maxTs[labelIndex])
        
        self.dropDownData["T Start"].clear()
        self.dropDownData["T Stop"].clear()
        self.addTValues(maxT)

    def addParentComboBox(self, parentComboBox, name = "Channel", addNA = False):
        channelSelect = QComboBox()
        channelSelect.setFixedWidth(200)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        self.formLayout.addRow(self.produceLabel(name), channelSelect)
        return channelSelect

    def produceLabel(self, text):
        label = QLabel(text)
        return label
    
    def getData(self):
        dropDownData = {k:v.currentText() for k,v in self.dropDownData.items()}
        lineEditData = {k:float(v.text()) for k,v in self.lineEditData.items()}
        boolData = {k:v.isChecked() for k,v in self.checkBoxData.items()}
        return dropDownData | lineEditData | boolData

    def addTimeDropDown(self, name):
        timeSelect = QComboBox()
        timeSelect.setFixedWidth(120)
        self.formLayout.addRow(self.produceLabel(name), timeSelect)
    
    def addTDropDown(self):
        hbox = QHBoxLayout()
        timeSelectStart, timeSelectStop = QComboBox(), QComboBox()
        timeSelectStart.setFixedWidth(120)
        timeSelectStart.setToolTip("Start time point")
        timeSelectStop.setFixedWidth(120)
        timeSelectStop.setToolTip("Stop Time Point (Inclusive)")
        hbox.addWidget(timeSelectStart)
        hbox.addWidget(timeSelectStop)
        self.dropDownData["T Start"] = timeSelectStart
        self.dropDownData["T Stop"] = timeSelectStop
        self.formLayout.addRow(QLabel("Time Start/Stop"), hbox)
    
    def addTValues(self, maxT, prefix = "T "):
        validTs = list(range(maxT+1))
        validTs = [str(t) for t in validTs]
        self.dropDownData[f"{prefix}Start"].addItems(validTs)
        self.dropDownData[f"{prefix}Start"].setCurrentIndex(0)
        self.dropDownData[f"{prefix}Stop"].addItems(validTs)
        self.dropDownData[f"{prefix}Stop"].setCurrentIndex(len(validTs)-1)

    def showErrorMessage(self):
        if not self.errorAlrShown:
            self.message.setText(self.message.text() + "\nEnsure T Start<T Stop and All Line Entries are Numeric")
            self.errorAlrShown = True

    def accept(self):
        if self.validateLineData() and self.validateTData():
            super().accept()
        else:
            self.showErrorMessage()
    
    def validateTData(self):
        data = self.dropDownData
        valid = int(data["T Start"].currentText())<=int(data["T Stop"].currentText())
        return valid

    def validateLineData(self):
        data = {k:v.text() for k,v in self.lineEditData.items()}
        i = 0
        vals = list(data.values())
        valid = True
        while i<len(vals) and valid:
            valid = self.isNumber(vals[i]) and float(vals[i])>=0.0
            i+=1
        return valid
    
    def isNumber(self, num):
        """ Returns True is string is a number. """
        # must remove periods to ensure decimals result in True
        return num.replace(".","").isdigit()

class ArtilifeParamDialog(ModelParamDialog):
    def __init__(self, params, paramtypes, name, parent):
        super().__init__(params,paramtypes,"artilife", parent)
    
    def populateFormLayout(self):
        self.dropDownData  = {}
        
        self.addArtilifeWeightDropDown()

        self.dropDownData["Channel"] =  self.addParentComboBox(self.parent.channelSelect, name = "Channel to Segment")
        # order is important here as self.dropDownData["T Start"]/["T Stop"]
            # must exist before addTValues and indexChange can be called
        self.addTDropDown()
        self.dropDownData["Channel"].currentIndexChanged.connect(self.channelSelectIndexChanged)
        if self.dropDowns:
            for dropDownName in self.dropDowns.keys():
                self.dropDownData[dropDownName] = self.addParentComboBox(self.parent.channelSelect, name = dropDownName, addNA = True)


        formDict = self.lineEdits
        self.lineEditData = {}
        for labelName,defaultVal in formDict.items():
            lineEdit  = QLineEdit(self)
            lineEdit.setFixedWidth(120)
            lineEdit.setText(str(defaultVal))
            self.lineEditData[labelName] = lineEdit
            self.formLayout.addRow(self.produceLabel(labelName), lineEdit)
    
        self.spinBoxData = {}
        for boxName, defaultVal in self.spinBoxes.items():
            spinBox = QSpinBox(self)
            spinBox.setValue(int(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        for boxName, defaultVal in self.dSpinBoxes.items():
            spinBox = QDoubleSpinBox(self)
            spinBox.setMaximum(1000.0)
            spinBox.setValue(float(defaultVal))
            self.spinBoxData[boxName] = spinBox
            self.formLayout.addRow(self.produceLabel(boxName), spinBox)
        
        self.addArtilifeTDropDowns(name = "mat")
        self.addArtilifeTDropDowns(name = "spore")
        self.channelSelectIndexChanged()
    
        # check boxes
        hLayout = QHBoxLayout()
        self.checkBoxData = {}
        if self.checkBoxes:
            for checkName, defaultVal in self.checkBoxes.items():
                checkBox = QPushButton(checkName)
                checkBox.setCheckable(True)
                hLayout.addWidget(checkBox)
                checkBox.setChecked(defaultVal)
                self.checkBoxData[checkName] = checkBox
            self.formLayout.addRow(hLayout)
        
        self.connectSporeMatButtons()

    def connectSporeMatButtons(self):
        self.checkBoxData["Mating Cells"].toggled.connect(self.toggleMat)
        self.checkBoxData["Sporulating Cells"].toggled.connect(self.toggleSpore)
    
    def toggleMat(self):
        enabled = self.checkBoxData["Mating Cells"].isChecked()
        self.dropDownData["matStart"].setEnabled(enabled)
        self.dropDownData["matStop"].setEnabled(enabled)
        self.dropDownData["matSeg"].setEnabled(enabled)
    def toggleSpore(self):
        enabled = self.checkBoxData["Sporulating Cells"].isChecked()
        self.dropDownData["sporeStart"].setEnabled(enabled)
        self.dropDownData["sporeStop"].setEnabled(enabled)
        self.dropDownData["tetradSeg"].setEnabled(enabled)

    def addArtilifeWeightDropDown(self):
        artilifeWeights = self.parent.getModelWeights(name = "artilife")
        artiWeights = QComboBox()
        artiWeights.addItems(artilifeWeights)
        self.dropDownData["artiWeights"] = artiWeights
        artiWeights.setFixedWidth(200)
        artiWeights.setEnabled(True)
        artiWeights.setCurrentIndex(0)
        self.formLayout.addRow(QLabel("artilife weights"), artiWeights)

    def addArtilifeTDropDowns(self, name):
        hbox = QHBoxLayout()
        timeSelectStart, timeSelectStop = QComboBox(), QComboBox()
        timeSelectStart.setEnabled(False)
        timeSelectStop.setEnabled(False)
        timeSelectStart.setFixedWidth(120)
        timeSelectStart.setToolTip("Start time point")
        timeSelectStop.setFixedWidth(120)
        timeSelectStop.setToolTip("Stop Time Point (Inclusive)")
        hbox.addWidget(timeSelectStart)
        hbox.addWidget(timeSelectStop)
        self.dropDownData[f"{name}Start"] = timeSelectStart
        self.dropDownData[f"{name}Stop"] = timeSelectStop
        longName = "Mating" if name == "mat" else "Sporulating"
        self.formLayout.addRow(QLabel(f"{longName} Start/Stop"), hbox)
        
        # add dropdown to allow for 
        modelType = "matSeg" if name == "mat" else "tetradSeg"
        modeloptions = [model for model in self.parent.modelNames if modelType in model]
        weights = QComboBox()
        weights.setEnabled(False)
        weights.setFixedWidth(120)
        weights.setToolTip(f"Weights for {modelType}")
        weights.addItems(modeloptions)
        weights.setCurrentIndex(0)
        self.dropDownData[modelType] = weights
        self.formLayout.addRow(QLabel(f"{modelType} Weights"), weights)


    def channelSelectIndexChanged(self):
        channelIndex = self.parent.channelSelect.findText(self.dropDownData["Channel"].currentText())
        maxT = self.parent.experiment().channels[channelIndex].max_t()
        for prefix in ["T ", "mat", "spore"]:
            self.dropDownData[f"{prefix}Start"].clear()
            self.dropDownData[f"{prefix}Stop"].clear()
            self.addTValues(maxT, prefix = prefix)

class TrainWindow(QDialog):
    def __init__(self, modelType, weightPath, parent):
        super().__init__(parent)
        self.setGeometry(100,100,900,350)
        self.setWindowTitle("Train Settings")
        
        self.parent = parent
        self.modelType = modelType
        self.weightPath = weightPath
        self.modelName = self.getModelName(self.weightPath)

        self.win = QWidget(self)
        self.l0 = QGridLayout()
        self.win.setLayout(self.l0)

        self.trainingParamVals, self.lossName = self.getTrainingParams()

        self.yoff = 0
        qlabel = QLabel('Train Model w/ Images')
        qlabel.setFont(QtGui.QFont("Arial", 14, QtGui.QFont.Bold))
        qlabel.setAlignment(QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,2)

        # choose initial model
        self.yoff+=1
        self.modelChoose = QComboBox()
        #self.modelChoose.addItems(modelNames)
        self.modelChoose.addItems([self.modelName], ) 
        self.modelChoose.setFixedWidth(150)
        #self.modelChoose.setCurrentIndex(self.parent.modelChoose.currentIndex())
        self.l0.addWidget(self.modelChoose, self.yoff, 1,1,1)
        qlabel = QLabel('initial model: ')
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,1)

        # add check box for from scratch training
        self.yoff+=1
        qlabel = QLabel(f"From Scratch")
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.checkbox = QCheckBox()
        self.l0.addWidget(qlabel, self.yoff,0,1,1)
        self.l0.addWidget(self.checkbox,self.yoff,1,1,1)

        # add loss function
        self.yoff+=1
        lossDisplay = QComboBox()
        lossDisplay.addItem(str(self.lossName))
        lossDisplay.setFixedWidth(150)
        self.l0.addWidget(lossDisplay, self.yoff,1,1,1)
        qlabel = QLabel("Loss Function: ")
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,1)

        # # select channels for training
        # imSelect = self.addImSelect("Images", self.parent.channelSelect)
        # labelSelect = self.addImSelect("Labels", self.parent.labelSelect)
        # for cSelect in [imSelect, labelSelect]:
        #     self.yoff+=1
        #     cSelect.setFixedWidth(150)
        #     self.l0.addWidget(cSelect.labelObj, self.yoff, 0,1,1)
        #     self.l0.addWidget(cSelect, self.yoff, 1,1,1)

        self.yoff+=1
        qlabel = QLabel("New Model Suffix")
        qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.l0.addWidget(qlabel, self.yoff,0,1,1)
        self.suffixName = QLineEdit()
        self.suffixName.setToolTip("Initial model name will automatically be added as a prefix to New Model Suffix.")
        self.suffixName.setText(self.parent.trainModelSuffix)
        self.suffixName.setFixedWidth(200)
        self.l0.addWidget(self.suffixName, self.yoff, 1,1,1)

        self.addTrainParams()
        
        self.addImNames()

        # click button
        self.yoff+=3
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.l0.addWidget(self.buttonBox, self.yoff, 0, 1,4)
        
    
    def accept(self):
        if self.validateTrainingParamVals():
            super().accept()
        else:
            self.showErrorMessage()
    
    def validateTrainingParamVals(self):
        data = self.edits.values()
        for entry in data:
            if not self.isNumber(entry.text()):
                return False
        return True
    
    def getData(self):
        suffix = str(self.suffixName.text())
        return {k:float(v.text()) for k,v in self.edits.items()} | {"scratch": self.checkbox.isChecked()} | {"model_name": f"{self.modelName}_{suffix}"}

    
    def isNumber(self, num):
        """ Returns True is stsring is a number. """
        # must remove periods to ensure decimals result in True
        return num.replace(".","").isdigit()

    def getModelName(self, weights):
        return os.path.split(weights)[-1].split(".")[0]

    def addImSelect(self, name, parentComboBox):
        parentComboBox = self.parent.channelSelect
        channelSelect = QComboBox()
        channelSelect.setFont(self.parent.medfont)
        channelSelect.setFixedWidth(120)
        channelSelect.addItems([parentComboBox.itemText(i) for i in range(parentComboBox.count())])
        channelSelect.setCurrentIndex(parentComboBox.currentIndex())
        channelSelect.labelObj = QLabel(name)
        return channelSelect
    
    def getTrainingParams(self):
        modelClass =  self.parent.getModelClass(self.modelType)
        modelParams = dict(modelClass.trainparams)
        loss = modelClass.loss
        return dict(modelParams), loss
    
    def addTrainParams(self):
        # choose parameters        
        labels = list(self.trainingParamVals.keys())
        defaultVals = list(self.trainingParamVals.values())
        self.edits = {}
        self.yoff += 1
        for i, label in enumerate(labels):
            qlabel = QLabel(label)
            qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            self.l0.addWidget(qlabel, i+self.yoff,0,1,1)
            newEdit = (QLineEdit())
            newEdit.setText(str(defaultVals[i]))
            newEdit.setFixedWidth(200)
            self.edits[label] = newEdit
            self.l0.addWidget(newEdit, i+self.yoff, 1,1,1)
    
    def addImNames(self):
         # list files in folder
        qlabel = QLabel('ims')
        qlabel.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Bold))
        self.l0.addWidget(qlabel, 0,4,1,1)
        qlabel = QLabel('ROIs')
        qlabel.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Bold))
        self.l0.addWidget(qlabel, 0,5,1,1)
        
        for i in range(10):
            if i > len(self.parent.trainFiles) - 1:
                break
            elif i==9 and len(self.parent.trainFiles) > 10:
                label = '...'
                nmasks = '...'
            else:
                label = os.path.split(self.parent.trainFiles[i])[-1]
                nmasks = str(self.parent.trainLabels[i].max())
            qlabel = QLabel(label)
            self.l0.addWidget(qlabel, i+1,4,1,1)
            qlabel = QLabel(nmasks)
            qlabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            self.l0.addWidget(qlabel, i+1, 5,1,1)



class PlotWindowCustomizeTimeSeries(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setGeometry(100,100,1000,450)
        self.setWindowTitle("Choose Properties to Plot Versus Time")
        self.win = QWidget(self)
        self.l = QGridLayout()
        self.win.setLayout(self.l)

        self.morphProps = self.parent.getCellDataLabelProps()
        self.trees = []
        self.boxes = {}
        self.produceParamTreeSingleCell()
        self.produceParamTreeIncludePopulations()

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.l.addWidget(self.buttonBox, 3, 0, 1, 2)

        self.win.show()
    
    def getProps(self, populationObj, propType):
        if propType == "morphology":
            return self.morphProps
        else:
            return [prop for prop in populationObj.properties if prop not in self.morphProps]

    def produceParamTreeSingleCell(self):
        populations = self.parent.getLabelsWithPopulationData()    
        
        tree = self.produceTree()
        self.trees.append(tree)
        level1 = "single"
        item = QTreeWidgetItem(tree, [level1])
        item.setExpanded(True)

        for level2 in populations:
            setName = self.parent.getTimeSeriesDataName(level2)
            item2 = QTreeWidgetItem(item, [setName])
            item2.setExpanded(True)

            for level3 in ["morphology", "pixel"]:
                item3 = QTreeWidgetItem(item2, [level3])

                for level4 in self.getProps(level2, level3):
                    item4 = QTreeWidgetItem(item3, [level4])
                    box = QCheckBox()
                    tree.setItemWidget(item4, 1, box)
                    self.boxes[PlotProperty(self.parent, "single", setName, level4)] = box
        self.l.addWidget(tree, 0,0,2,1)

    def produceParamTreeIncludePopulations(self):
        populations = self.parent.getLabelsWithPopulationData()      
        for i,level1 in enumerate(self.parent.populationPlotTypes):
            tree = self.produceTree()
            self.trees.append(tree)
            item = QTreeWidgetItem(tree, [level1])
            item.setExpanded(True)

            for level2 in populations:
                setName = self.parent.getTimeSeriesDataName(level2)
                item2 = QTreeWidgetItem(item, [setName])
                item2.setExpanded(True)

                for level2a in level2.population_names:
                    item2a = QTreeWidgetItem(item2, [level2a])

                    for level3 in ["morphology", "pixel"]:
                        item3 = QTreeWidgetItem(item2a, [level3])

                        for level4 in self.getProps(level2, level3):
                            item4 = QTreeWidgetItem(item3, [level4])
                            box = QCheckBox()
                            tree.setItemWidget(item4, 1, box)
                            self.boxes[PlotProperty(self.parent, level1, setName, level4, populationName=level2a)] = box
            self.l.addWidget(tree, 0,i+1,2,1)

    def getData(self):
        checked = [name for name, box in self.boxes.items() if box.isChecked()]
        toPlotDict = {}
        for plot in self.parent.plotTypes:
            toPlotDict[plot] = [prop_obj for prop_obj in checked if plot in prop_obj.plotType]
        return toPlotDict

    
    def produceTree(self):
        tree = QTreeWidget()
        tree.setColumnCount(2)
        tree.setHeaderHidden(True)
        tree.setMinimumHeight(300)
        tree.setMinimumWidth(300)
        tree.header().setSectionResizeMode(QHeaderView.ResizeToContents)
        return tree


class PlotWindowCustomizePerFrame(QDialog):
    def __init__(self, properties, masks, max_selected=6, parent=None):
        super(PlotWindowCustomizePerFrame, self).__init__(parent)
        self.properties = properties
        self.masks = masks
        self.max_selected = max_selected
        self.selected_properties = {mask: [] for mask in masks}
        self.init_ui()
        
    def init_ui(self):
        main_layout = QVBoxLayout()  # Main layout for the dialog
        
        self.layout = QHBoxLayout()  # Horizontal layout for the checkboxes
        self.checkboxes = {mask: [] for mask in self.masks}
        for mask in self.masks:
            mask_layout = QVBoxLayout()
            mask_label = QLabel(mask)
            mask_layout.addWidget(mask_label)
            for prop in self.properties:
                checkbox = QCheckBox(prop)
                checkbox.stateChanged.connect(self.on_checkbox_state_changed)
                mask_layout.addWidget(checkbox)
                self.checkboxes[mask].append(checkbox)
            self.layout.addLayout(mask_layout)
        
        main_layout.addLayout(self.layout)

        # Add mutually exclusive checkboxes for frames
        self.frame_checkboxes = QHBoxLayout()
        self.calculate_all_frames_checkbox = QCheckBox("Calculate for all frames")
        self.calculate_current_frame_checkbox = QCheckBox("Calculate for current frame")
        self.calculate_current_frame_checkbox.setChecked(True)
        
        self.frame_button_group = QButtonGroup()
        self.frame_button_group.addButton(self.calculate_all_frames_checkbox)
        self.frame_button_group.addButton(self.calculate_current_frame_checkbox)
        self.frame_button_group.setExclusive(True)
        
        self.frame_checkboxes.addWidget(self.calculate_all_frames_checkbox)
        self.frame_checkboxes.addWidget(self.calculate_current_frame_checkbox)
        
        main_layout.addLayout(self.frame_checkboxes)
        
        # Add OK button
        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok)
        self.button_box.accepted.connect(self.accept)
        main_layout.addWidget(self.button_box)
        
        self.setLayout(main_layout)
        self.setWindowTitle("Select Properties")
        
    def on_checkbox_state_changed(self, state):
        total_selected = sum(len(props) for props in self.selected_properties.values())

        sender = self.sender()
        
        if state == Qt.Checked and total_selected >= self.max_selected:
            QMessageBox.warning(self, "Limit Exceeded", f"A maximum of {self.max_selected} properties can be selected.")
            sender.blockSignals(True)
            sender.setChecked(False)
            sender.blockSignals(False)
            return
        
        for mask, checkboxes in self.checkboxes.items():
            self.selected_properties[mask] = [cb.text() for cb in checkboxes if cb.isChecked()]
    
    def get_data(self):
        return self.selected_properties
    
    def get_unique_properties(self):
        unique_properties = set()
        for props in self.selected_properties.values():
            unique_properties.update(props)
        return list(unique_properties)
    
    def do_for_all_frames(self):
        return self.calculate_all_frames_checkbox.isChecked()





from yeastvision.models.matSeg.model import MatSeg

class SpoSeg(MatSeg):
    hyperparams  = { 
    "mean_diameter":30,
    "flow_threshold":0.4, 
    "cell_probability_threshold": 0.0,
    "half_image_size": True}
    types = [float, float, float, bool]
    def __init__(self, params, weights):
        super().__init__(params, weights)



from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtWidgets import (QApplication, QRadioButton, QButtonGroup, QSlider, QStyle, QStyleOptionSlider, QMainWindow, QPushButton, QDialog,
                            QDialogButtonBox, QLineEdit, QFormLayout, QMessageBox,
                             QFileDialog, QVBoxLayout, QCheckBox, QFrame, QSpinBox, QLabel, QWidget, QComboBox, QTableWidget, QSizePolicy, QGridLayout, QHBoxLayout)
import pyqtgraph as pg
from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea
from yeastvision.plot.cell_table import TableModel
from yeastvision.plot.types import SingleCellUpdatePlot, HeatMap, PopulationPlot, SingleFrameUpdatePlot
from yeastvision.track.data import PopulationReplicate, Population
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.colors import LinearSegmentedColormap, BoundaryNorm

class HeatmapCanvas(FigureCanvas):
    def __init__(self, intervals, max_frame, interp_intervals, parent=None):
        fig, self.ax = plt.subplots(figsize=(10, 4))  # Increased figure size for better layout
        super().__init__(fig)
        self.setParent(parent)

        self.create_1d_heatmap(intervals, max_frame, interp_intervals)

    def create_1d_heatmap(self, intervals, max_frame, interp_intervals):
        # Create an array for the heatmap
        heatmap = np.zeros(max_frame)

        for interval in interp_intervals:
            start = interval['start']
            stop = interval['stop']
            interp = int(interval['interp'])
            interp_factor = 2 ** interp

            i = start
            while i < stop:
                heatmap[i+1:(i+1) + (interp_factor-1)] = interp
                i = (i+1) + (interp_factor-1)
        #print("Heatmap values:", heatmap)  # Debugging line to check heatmap values

        # Create a modern-looking colormap
        colors = [
            (1, 1, 1, 0),       # Transparent for 0 interpolation
            (0.8, 0.8, 1, 1),   # Light blue for 2x
            (0.6, 0.6, 1, 1),   # Medium blue for 4x
            (0.4, 0.4, 1, 1),   # Dark blue for 8x
            (0.2, 0.2, 1, 1),   # Darker blue for 16x
        ]
        cmap = LinearSegmentedColormap.from_list("interpolation_cmap", colors)
        norm = BoundaryNorm([0, 1, 2, 3, 4, 5], cmap.N)

        # Create the 1D heatmap
        heatmap_image = heatmap.reshape(1, -1)

        im = self.ax.imshow(heatmap_image, aspect='auto', cmap=cmap, norm=norm, extent=[0, max_frame, 0, 1])

        # Customize the plot
        self.ax.set_yticks([])
        self.ax.set_xlabel('Frame')
        self.ax.set_title('Interpolation Levels')

        # Draw lines beneath the graph for each interval
        for interval in interp_intervals:
            start = interval['start']
            stop = interval['stop']
            interp = int(interval['interp'])
            interp_factor = 2 ** interp

            self.ax.hlines(-0.1, start, stop, colors=cmap(norm(interp)), lw=4)
            self.ax.vlines(start, -0.15, -0.05, colors='black', lw=1)
            self.ax.vlines(stop, -0.15, -0.05, colors='black', lw=1)

        # Add a custom horizontal legend below the x-axis
        handles = [
            plt.Line2D([0], [0], color=cmap(norm(1)), lw=4, label='2x'),
            plt.Line2D([0], [0], color=cmap(norm(2)), lw=4, label='4x'),
            plt.Line2D([0], [0], color=cmap(norm(3)), lw=4, label='8x'),
            plt.Line2D([0], [0], color=cmap(norm(4)), lw=4, label='16x')
        ]
        legend = self.ax.legend(handles=handles, loc='upper center', bbox_to_anchor=(0.5, -0.3),
                                ncol=4, title="Interpolation Levels")

        plt.tight_layout(pad=3.0)  # Ensure the layout fits properly
        self.draw()

class InterpolationHeatmapWindow(QWidget):
    def __init__(self, parent, intervals, max_frame, interp_intervals):
        super().__init__()
        self.parent = parent
        self.setWindowTitle("Interpolation Heatmap Viewer")

        layout = QVBoxLayout(self)

        label = QLabel(f"Interpolation Intervals (Original Movie) {str(intervals)}")
        layout.addWidget(label)

        self.heatmap_canvas = HeatmapCanvas(intervals, max_frame, interp_intervals, 
                                            parent = self)
        layout.addWidget(self.heatmap_canvas)


def plot_tree(matrix, selected_cells=[]):
    '''
    PARAMETERS
    -------------
    matrix: np.ndarray
        matrix that encodes lineage and cell info as follows:
            column 0: cell idx (sorted numerically, 1-indexed)
            column 1: cell birth frame (0 indexed)
            column 2: cell death frame (0 indexed)
            column 3: mother idx (-1 if has no mother)
        lineage tree places initial, seed cells on seperate frames
    
    selected cell: list (optional, default [])
        cells whose daughters should be highlighted
    

    RETURNS
    ---------------
    figure: matplotlib.pyplot.Figure
        can be embedded into a GUI window or shown with matplotlib.pylot.show
    
    '''
    fig, ax = plt.subplots(figsize=(10, 6))
    y_positions = {}
    current_y = 0  # Starting position

    # Bright color list for selected cells
    bright_colors = [
        '#FF0000',  # Bright red
        '#00FF00',  # Bright green
        '#0000FF',  # Bright blue
        '#FF00FF',  # Bright magenta
        '#00FFFF',  # Bright cyan
        '#FFFF00',  # Bright yellow
    ]

    default_horizontal_color = 'blue'
    default_vertical_color = 'red'

    if selected_cells:
        cell_colors = {selected_cells[i]: bright_colors[i % len(bright_colors)] for i in range(len(selected_cells))}
    else:
        cell_colors = {}

    def plot_lineage(mother, y):
        daughters = matrix[matrix[:, 3] == mother]
        daughters = daughters[daughters[:, 1].argsort()[::-1]]

        mother_color = cell_colors.get(mother, default_horizontal_color)

        for daughter in daughters:
            cell_index, birth_frame, death_frame, _ = daughter
            y += 1

            while y in y_positions.values():
                y += 1

            horizontal_color = mother_color if cell_index in cell_colors or mother_color != default_horizontal_color else default_horizontal_color
            vertical_color = mother_color if mother_color != default_horizontal_color else default_vertical_color
            
            ax.plot([birth_frame, death_frame], [y, y], color=horizontal_color)
            ax.text(death_frame, y, str(int(cell_index)), color=horizontal_color, verticalalignment='center',
                    horizontalalignment='left')
            ax.plot([birth_frame, birth_frame], [y_positions[mother], y], color=vertical_color)

            y_positions[cell_index] = y
            y = plot_lineage(cell_index, y)

        return y

    no_mothers = matrix[matrix[:, 3] == -1]
    for row in no_mothers:
        cell_index, birth_frame, death_frame, _ = row
        color = cell_colors.get(cell_index, default_horizontal_color)
        ax.plot([birth_frame, death_frame], [current_y, current_y], color=color)
        ax.text(death_frame, current_y, str(int(cell_index)), color=color, verticalalignment='center',
                horizontalalignment='left')
        y_positions[cell_index] = current_y
        current_y = plot_lineage(cell_index, current_y) + 1

    ax.set_ylim(-1, current_y + 1)
    ax.set_xlabel('Frame')
    ax.get_yaxis().set_visible(False)
    plt.grid(False)
    ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True))
    plt.tight_layout()
    return fig



class LineageTreeWindow(QWidget):
    def __init__(self, parent, matrix, selected_cells=[]):
        super().__init__()
        self.parent = parent
        self.matrix = matrix
        self.selected_cells = selected_cells
        self.initUI()

    def initUI(self):
        layout = QVBoxLayout()

        # Create the matplotlib canvas and add it to the layout
        fig = plot_tree(self.matrix, self.selected_cells)
        canvas = FigureCanvas(fig)
        layout.addWidget(canvas)

        # Optionally add more widgets like buttons to the layout
        # button = QPushButton('Close')
        # button.clicked.connect(self.close)
        # layout.addWidget(button)

        self.setLayout(layout)
        self.setWindowTitle('Cell Lineage Tree')
        self.show()

    def closeEvent(self, event):
        self.parent.showTreeButton.setCheckState(False)


class PlotProperty():
    def __init__(self, parent, plotType, setName, property, populationName = None):
        self.parent = parent
        self.plotType = plotType
        self.labelIdx = self.getLabelIdx(setName)
        self.prop = property
        self.population = populationName
    
    def getPlotName(self):
        labelName = self.getLabelName()
        if self.populationName:
            return "f{labelName}-{self.population}-{self.prop}"
        else:
            return "f{labelName}-{self.prop}"
    
    def getLabelIdx(self, setName):
        return self.parent.labelSelect.items().index(setName)
    
    def getLabelName(self):
        return self.labelSelect.items()[self.labelIdx]
    
    def get_timeseries_obj(self):
        return self.parent.experiment().labels[self.labelIdx].celldata
    
    @staticmethod
    def cluster_property(plot_property_list):
        by_property = {}
        for plot_property_obj in plot_property_list:
            prop = plot_property_obj.prop
            if prop not in by_property:
                by_property[prop] = [plot_property_obj]
            else:
                by_property[prop].append(plot_property_obj)
        return by_property

    @staticmethod
    def cluster_population(plot_property_list):
        by_population = {}
        for plot_property_obj in plot_property_list:
            pop = plot_property_obj.population
            if pop not in by_population:
                by_population[pop] = [plot_property_obj]
            else:
                by_population[pop].append(plot_property_obj)
        return by_population

    

class EvalWindow(QWidget):
    def __init__(self, parent, data_dict, IOUs):
        '''
        params
        data_dict - key: name of the plot (precision, recall, F1, accuracy), values - list of tuple for each set of labels being 
        evaluated where tup1 is the name and tup2 is data vs IOU score
        '''
        super(EvalWindow, self).__init__()
        self.setWindowTitle("Evaluation Window")

        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.data = data_dict

        self.numPlots = len(data_dict)
        self.evalNames = [n for n,_ in list(self.data.values())[0]]
        self.numToEval = len(self.evalNames)
        self.xVals = IOUs

        # Create a table widget to display data
        self.tableLayout = QVBoxLayout()
        self.table_widgets = []

        self.toolbar = NavigationToolbar(self.canvas, self)


        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        #layout.addWidget(self.tableLayout)

        self.setLayout(layout)

        self.setData()
        self.setGeometry(400,400,500,500)
        self.canvas.draw()
        self.show()

    def setData(self):
        i = 0
        for plotType, data in self.data.items():
            self.producePlot(data, plotType, i)
            i+=1

    
    # def produceTable(self, title, data):
    #     table_widget = QTableWidget()
    #     table_widget.setColumnCount(self.numToEval)
    #     table_widget.setRowCount(len(data[0]))

    #     for col, line_data in enumerate(data):
    #         for row, value in enumerate(line_data):
    #             item = QTableWidgetItem(str(value))
    #             table_widget.setItem(row, col, item)

    #     return table_widgets

    
    def producePlot(self, datas, title, col):
        ax = self.figure.add_subplot(2, 2, col + 1)
        for name, data in datas:
            ax.plot(self.xVals, data, 'o-', label = name)
            ax.set_aspect('auto')
            ax.set_xbound(lower = 0.45,upper = 1.05)
            ax.set_xticks([0.5,0.6,0.7,0.8,0.9,1])

            if col == 0:
                ax.legend()
                ax.set_xlabel("IOU Matching Threshold")
        ax.set_title(title)
        plt.tight_layout()

class SingleFramePlotWindow(QWidget):
    def __init__(self, parent, propDict, data, allFrames=False, plots_per_row=3):
        super().__init__()
        self.parent = parent
        self.propDict = propDict
        self.data = data
        self.allFrames = allFrames
        self.plots_per_row = plots_per_row
        self.docks = []
        self.plots = []  # Store the plot instances for updating
        self.initUI()

        if self.allFrames:
            self.update(self.parent.tIndex)

        #self.setGeometry(500, 100, 1500, 1500)
    
    def initUI(self):
        layout = QVBoxLayout()
        self.setLayout(layout)
        
        self.area = DockArea()
        layout.addWidget(self.area)
        
        self.createPlots()
    
    def createPlots(self):
        row = 0
        col = 0
        
        for mask, properties in self.propDict.items():
            for prop in properties:
                plot = SingleFrameUpdatePlot(self, prop, mask, self.data[mask])
                plot_widget = plot.getPlotWidget()
                
                dock_name = f"{mask}_{prop}"
                dock = Dock(dock_name, closable=True)
                dock.addWidget(plot_widget)
                
                if col == 0 and row == 0:
                    self.area.addDock(dock, 'top')
                else:
                    position = 'right' if col > 0 else 'bottom'
                    relative_to = self.docks[-1] if col > 0 else self.docks[-self.plots_per_row]
                    self.area.addDock(dock, position, relative_to)
                
                self.docks.append(dock)
                self.plots.append(plot)
                col += 1
                if col >= self.plots_per_row:
                    col = 0
                    row += 1
    
    def update(self, frame_index):
        for plot in self.plots:
            plot.update(frame_index)
    
    def closeEvent(self, event):
        self.parent.sfPlotWindowOn = False


class TimeSeriesPlotWindow(QWidget):
    def __init__(self, parent, propDict):
        super().__init__()
        self.parent = parent
        pg.setConfigOption('imageAxisOrder', 'row-major')
        self.hLayout = QHBoxLayout()
        self.setLayout(self.hLayout)
        self.createTable()
        pg.setConfigOption('imageAxisOrder', 'row-major')
        self.cmap = pg.colormap.get("jet", source = 'matplotlib')
        self.setData()
        
        self.propDict = propDict


        if self.hasPlots():
            self.area = DockArea()
            self.hLayout.addWidget(self.area, 7)
            self.docks = {} # key = name of dock, value = (dock object, widget contained in dock)
            if self.hasSingleCellPlots():
                self.addSingleCellPlots()
            if self.hasHeatmaps():
                self.addHeatmaps()
            if self.hasPopulationPlots():
                self.addPopulationsPlots()
            self.setGeometry(500, 100, 1500, 1500)
    
    def hasPlots(self):
        return self.hasSingleCellPlots() or self.hasHeatmaps() or self.hasPopulationPlots()

    def hasSingleCellPlots(self):
        return "single" in self.propDict and self.propDict["single"]
    
    def hasMplPlots(self):
        return self.hasPopulationPlots()
    
    def hasHeatmaps(self):
        return "heatmap" in self.propDict and self.propDict["heatmap"]
    
    def hasPopulationPlots(self):
        return "population" in self.propDict and self.propDict["population"]

    def updateAll(self):
        if self.hasSingleCellPlots():
            self.updateSingleCellPlots()
        if self.hasHeatmaps():
            self.updateHeatMaps()
        if self.hasPopulationPlots():
            self.updatePopulationPlots()

    def addSingleCellPlots(self):
        self.singleCellPlots = []
        single_cell_prop_obj = self.propDict["single"]
        clustered_by_property = PlotProperty.cluster_property(single_cell_prop_obj)
        for property in clustered_by_property:
            timeseries = [prop.get_timeseries_obj() for prop in clustered_by_property[property]]
            self.singleCellPlots.append(SingleCellUpdatePlot(self.parent, property, timeseries, self.parent.selectedCells))
        
 
        dockName = f"single plots"
        newDock = Dock(dockName, closable = True)
        dockLayout = DockArea()
        newDock.addWidget(dockLayout)
        self.area.addDock(newDock, "bottom")
        

        for i, plot in enumerate(self.singleCellPlots):
            if i == 0:
                position = "bottom"
                relativeTo = None
            else:
                position = "above"
                plotRelativeTo = self.singleCellPlots[i-1]
                relativeTo = self.docks[plotRelativeTo]

            name = f"{plot.property}"
            plotDock = Dock(name, closable = True)
            plotDock.addWidget(plot.getPlotWidget())
            self.docks[plot] = (plotDock)
  
            dockLayout.addDock(plotDock, position, relativeTo)
    
    def addHeatmaps(self):
        self.heatMaps = []
        heatmap_prop_obj = self.propDict["heatmap"]
        clustered_by_property = PlotProperty.cluster_property(heatmap_prop_obj)
        for property, properties in clustered_by_property.items():
            timeseries = [prop.get_timeseries_obj() for prop in properties]
            populations = [prop.population for prop in properties]
            for ts, population in zip(timeseries, populations):
                self.heatMaps.append(HeatMap(self, property, population, ts))
    
        dockName = f"heatmaps"
        newDock = Dock(dockName, closable = True)
        dockLayout = DockArea()
        newDock.addWidget(dockLayout)
        self.area.addDock(newDock, "bottom")

        for i, plot in enumerate(self.heatMaps):
            if i == 0:
                position = "bottom"
                relativeTo = None
            else:
                position = "above"
                plotRelativeTo = self.heatMaps[i-1]
                relativeTo = self.docks[plotRelativeTo]

            name = f"{plot.property}"
            plotDock = Dock(name, closable = True)
            plotDock.addWidget(plot.getPlotWidget())
            self.docks[plot] = (plotDock)

            dockLayout.addDock(plotDock, position, relativeTo)
    
    def addPopulationsPlots(self):
        self.populationPlots = []
        population_prop_obj = self.propDict["population"]
        clustered_by_property = PlotProperty.cluster_property(population_prop_obj)
        for property, properties in clustered_by_property.items():
            timeseries = [prop.get_timeseries_obj() for prop in properties]
            populations = [prop.population for prop in properties]
            for ts, population in zip(timeseries, populations):
                self.populationPlots.append(PopulationPlot(self, property, ts, population))
    
        dockName = f"population plots"
        newDock = Dock(dockName, closable = True)
        dockLayout = DockArea()
        newDock.addWidget(dockLayout)
        self.area.addDock(newDock, "bottom")

        for i, plot in enumerate(self.populationPlots):
            if i == 0:
                position = "bottom"
                relativeTo = None
            else:
                position = "above"
                plotRelativeTo = self.populationPlots[i-1]
                relativeTo = self.docks[plotRelativeTo]

            name = f"{plot.property}"
            plotDock = Dock(name, closable = True)
            plotDock.addWidget(plot.getPlotWidget())
            self.docks[plot] = (plotDock)

            dockLayout.addDock(plotDock, position, relativeTo)
    
    def addMplPlots():
        pass

    def updateSingleCellPlots(self):
        for plot in self.singleCellPlots:
            plot.update()
        

    def updatePopulationPlots(self):
        for plot in self.populationPlots:
            plot.update()
    
    def updateHeatMaps(self):
        for plot in self.heatMaps:
            plot.update()

    def setData(self):
        self.data = self.parent.getCellData()

    def getDataIdx(self, population):
        try:
            idx = self.parent.labelSelect.items().index(population)
            return idx
        except ValueError:
            return None

    def updatePlots(self):
        self.updateSingleCellPlots()
        self.updatePopulationPlots()
        self.updateHeatMaps()

    def createTable(self):
        # create table with dummy data
        self.table = QtWidgets.QTableView()
        self.model = TableModel(self.parent.getCellDataAbbrev(), self.parent)
        self.table.setModel(self.model)

        self.table.selectionModel().currentChanged.connect(self.model.handleselect)

        # configure the table appearance
        self.table.verticalHeader().setVisible(False) # turn off vertical header
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch) # stretch last section to take full width of allottedd area
        self.table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.table.resizeColumnsToContents()
        self.hLayout.addWidget(self.table, 4)
        
    def resizeTableToContents(self):
        vh = self.table.verticalHeader()
        hh = self.table.horizontalHeader()
        size = QtCore.QSize(hh.length(), vh.length())  # Get the length of the headers along each axis.
        size += QtCore.QSize(vh.size().width(), hh.size().height())  # Add on the lengths from the *other* header
        size += QtCore.QSize(20, 20)  # Extend further so scrollbars aren't shown.
        self.resize(size)
    
    def closeEvent(self, event):
        self.parent.tsPlotWindowOn = False


from skimage.measure import regionprops, regionprops_table
from skimage.measure import shannon_entropy as shentr
from skimage.measure import label
import numpy as np
import pandas as pd
from yeastvision.utils import normalize_im
from tqdm import tqdm
from typing import List

def average_axis_lengths(object_array):
    """
    Given a 2D NumPy array with labeled objects, compute the average axis_major_length
    and axis_minor_length for all objects.
    
    Parameters:
    object_array (np.ndarray): A 2D NumPy array where each object is labeled with a unique integer.
    
    Returns:
    tuple: A tuple containing the average major axis length and average minor axis length.
    """
    
    # Get region properties for each labeled object
    labeled_objects = label(object_array)
    regions = regionprops(labeled_objects)
    
    # Collect the major and minor axis lengths for each object
    major_lengths = [region.axis_major_length for region in regions]
    minor_lengths = [region.axis_minor_length for region in regions]
    
    # Calculate the average major and minor axis lengths
    avg_major_length = np.mean(major_lengths) if major_lengths else 0
    avg_minor_length = np.mean(minor_lengths) if minor_lengths else 0
    
    return avg_major_length, avg_minor_length

def getBirthFrame(trackedMasks, cellVal):
    return np.min((np.nonzero(trackedMasks == cellVal))[0])

def getDeathFrame(trackedMasks, cellVal):
    return np.max((np.nonzero(trackedMasks == cellVal))[0])

def entropy(label, intensity_im):
    return shentr(intensity_im, base = 2)

def var(label, intensity_im):
    return np.var(intensity_im)

def median(label, intensity_im):
    return np.median(intensity_im)

def intensity_total(label, intensity_im):
    return np.sum(intensity_im)

def concentration(label, intensity_im):
    return intensity_total(label, intensity_im)/np.sum(label)

LABEL_PROPS = ["area", "eccentricity", "label", ]
IM_PROPS = ["intensity_mean"]
EXTRA_IM_PROPS = [var, intensity_total, concentration]

def getLifeData(labels):
    lifeData = {"birth":[], "death":[]}
    for val in tqdm(np.unique(labels)[1:]):
        lifeData["birth"].append(getBirthFrame(labels, val))
        lifeData["death"].append(getDeathFrame(labels, val))
    return lifeData

def getPropDf(data):
    #newData = data.drop[data[data["labels"] == "population"].index]
    try:
        return data.drop(columns = ["labels", "birth", "death"])
    except:
        return data.drop(columns = ["labels"])

def exportCellData(data, lifedata):
    export = {"cell":[], "time":[]}
    propDf = getPropDf(data)
    for column in propDf.columns:
            export[column] = []
    
    for cell in data["labels"]:
        firstT = int(lifedata[lifedata["cell"] == cell]["birth"])
        endT = int(lifedata[lifedata["cell"] == cell]["death"])

        for i in range(firstT, endT+1):
            export["time"].append(i)
            export["cell"].append(cell)

        for props in propDf.columns:
            vals = data[data["labels"] == cell]
            vals = vals[props].tolist()[0][firstT:endT+1]
            for prop in vals:
                export[props].append(prop)
    return pd.DataFrame(data = export)

def getDaughterMatrix(self, lineageDF):
    arraysize = len(lineageDF.index.tolist())
    daughterMatrix = np.zeros((arraysize, arraysize))
    return daughterMatrix

def getPotentialHeatMapNames(data):
    return getPropDf(data).columns


def getHeatMaps(data, chosen):
    heatMaps = {}
    for prop in chosen:
        hm = data[prop].tolist()
        heatMaps[prop] = (np.array(hm))
    return heatMaps

def getCellData(labels: np.ndarray, intensity_ims: List[np.ndarray] = None, intensity_im_names:List[str] = None)->pd.DataFrame:
    '''
    Computes cell data relating to image and morphological features of the given labels. If intensity images
    are provided, then they are used for image feature extraction (ie signal intensity, mean, concentration) for
    the objects in labels. 
    
    Args
        labels (ndarray): a t x h x w array with t timepoints of h x w images containing tracked objects, identified
                        by unique indeces. Any unsigned integer type is acceptable (uint8, uint16, uint32, etc)
        intensity_ims (optional, list of ndarrays): a list of images, each of which must have the same shape as `labels`. Each will
                                                    be used to compute pixel features for each object in `labels`. Each 3D time series from
                                                    this list be normalized to the interval [0,1]
        inensity_im_names (list of strings, optional): the string identifiers of each element of `intensity_ims`. Used
                                                    to identify each pixel feature in the resulting celldata DataFrame. If none,
                                                    channel1, channel2,...,channeln is the default naming scheme
    Returns:
        pd.DataFrame - a container storing the cell properties (columns) over time for given a object index in labels (rows). Each
                        entry is a t-vector containing time series for all cell-property pairs

    '''
    total_labels = np.unique(labels) # all indeces of cells present in the movie
    total_labels = total_labels[total_labels!=0]

    total_data = {"labels": total_labels} #return value, dictionary storing cell data, to be converted to pd.df
                                          # Will have keys representing properties and values as list of lists (ie list of cell
                                         # timeseries)
    
    # add birth and death data
    total_data = total_data #| getLifeData(labels)

    # determine whether or not to include image properties in call to regionprops_table
    if intensity_ims:
        intensity_ims = [[normalize_im(im) for im in intensity_im] for intensity_im in intensity_ims]
        props = LABEL_PROPS + IM_PROPS
        extra_properties = EXTRA_IM_PROPS
    else:
        props = LABEL_PROPS
        extra_properties = None

    # COMPUTE IMAGE DATA PER TIMEPOINT and reformat into output dictioary
    # Prepare data into a format for skimage.measure.regionprops_table
    for i in tqdm(range(labels.shape[0])):
        label = labels[i,:,:]

        # The intensity_im input to regionprops must be extracted stacked on the third dimensios
        if intensity_ims:
            if len(intensity_ims)>1:
                intensity_im = np.stack([im[i] for im in intensity_ims], -1)
            else:
                intensity_im = intensity_ims[0][i]
        else:
            intensity_im = None
        
        # get imagedata
        im_data = regionprops_table(label, intensity_im, properties = props, extra_properties = extra_properties)
        
        # CELL BIRTH AND DEATH CORRECTIONN: if a label is missing at a given timepoint,
        # then we must insert a None into im_data at the location of 
        # add zeros for cells not present in im data
        for label in total_labels:
            if label not in im_data["label"]:
                # an index is not present at a certain timepoit
                for property in im_data:
                    vals = [round(val, 3) if val is not None else None for val in list(im_data[property])]
                    im_data[property] = vals

                    if property != "label":
                        # prevent index out of bound error
                        if label-1 < len(im_data[property]):
                            im_data[property][label-1] = None
                        else:
                            im_data[property].append(None)
        
        # for every property that isnt "labels" in im_data, we store a list of lists
        # into total data under the key representing the corresponding property
        for property in im_data:
            if property != "label":

                # check if key exists to determine insertion format 
                if property not in total_data:
                    total_data[property] = [[prop] for prop in im_data[property]]
                else:
                    for label in total_labels:
                        total_data[property][label-1].append(im_data[property][label-1])
    
    # Get intensity image with correct names
    if intensity_im_names:
        # determine which properteis are pixel features
        for property in im_data:

            # determine if property is a pixel (image) feature
            isImProp = False
            for prop in IM_PROPS+EXTRA_IM_PROPS:
                try:
                    prop = prop.__name__
                except AttributeError:
                    prop = str(prop)
                
                if prop in property:
                    isImProp = True
                    break
            # found an image prop, break the loop and exit
            if isImProp:
                if "-" in property:
                    idx = int(property.split("-")[-1])-1
                else:
                    idx = 0
                # swap new name, be sure to delete entry under old name
                newPropertyName = "-".join([intensity_im_names[idx], prop])
                total_data[newPropertyName] = total_data[property]
                del total_data[property]

    # get population average for each property
    for property in total_data:
        if property not in ["labels", "birth", "death"]:
            property_array = np.array(total_data[property], dtype = np.float32) # rows are cells (axis 0), columns are time (axis 1)
            property_array[property_array == 0.0] = np.nan
            total_data[property].append(list(np.nanmean(property_array, axis = 0)))
        elif property == "labels":
            total_data["labels"] = list(total_data["labels"])
            total_data["labels"].append("population")
        else:
            total_data[property].append(0)

    df = pd.DataFrame.from_dict(total_data)
    return df



import numpy as np
from typing import List, Optional
import math
from skimage.transform import resize
from skimage.morphology import skeletonize, erosion, square
from skimage.filters import threshold_otsu
from scipy.stats import mode
from yeastvision.track.fiest.utils import cal_allob2, cal_allob1, cal_celldata, replace_none_with_empty_array, binar

def track_tetrads(spoSeg_output:np.ndarray, tetrad_interval: List, movie_length:int, shock_period:Optional[List[int]]=None):
    # define some global variables
    thresh_percent = 0.015
    thresh_remove_last_mask = 10
    thresh_next_cell = 400
    thresh = 80

    # tetrad masks are extended to go from index 0 of the movie to the end of the tetrad interval
    tet_masks = [None] * tetrad_interval[-1]
    im_shape = spoSeg_output[0].shape
    for i in range(len(tet_masks)): 
        if i >= tetrad_interval[0]:
            tet_masks[i] = spoSeg_output[i]
        else:
            tet_masks[i] = np.zeros(im_shape, dtype=np.uint16)
    
    for start, end in [shock_period]:
        for i in range(start, end):
            tet_masks[i] = None
    
    ###  find timepoint of first TET detection
    start = -1
    for its in range(len(tet_masks)):
        if tet_masks[its] is not None and np.sum(tet_masks[its]) > 0:
            start = its
            break

    ### determine the period to track or rang
    if start != -1:
        rang = range(start, len(tet_masks))
        I2 = tet_masks[start]
        A = np.zeros_like(tet_masks[start])
    else:
        rang = range(len(tet_masks))
        I2 = tet_masks[0]
        A = np.zeros_like(tet_masks[0])


    # Tracking all detections
    #  actual tracking loop

    IS6 = np.zeros_like(I2)
    TETC = [None] * 2
    TETC[0] = [None] * len(tet_masks)
    TETC[1] = [None] * len(tet_masks)
    xx = start
    rang2 = rang
    ccel = 1

    while xx != -1:
        k = 0
        for im_no in rang2:
            # im_no = 72
            I2 = tet_masks[im_no] if ccel == 1 else TETC[1][im_no]
            if I2 is None or I2.size == 0:
                continue
            if im_no == min(rang2):
                ind1 = np.unique(I2)[1:]  # Exclude background
                I3 = (I2 == ind1[0])
                I3A = I3.copy()
            else:
                I3A = IS6.copy()

            I3A = skeletonize(binar(I3A))
            I2A = I2.copy()
            I3B = I3A.astype(np.uint16) * I2A.astype(np.uint16)
            ind = mode(I3B[I3B != 0])[0]

            if (ind == 0 or math.isnan(ind)) and ccel == 1:
                k += 1
                if k > thresh_next_cell:
                    for im_no_1 in range(im_no, rang[-1] + 1):
                        if tet_masks[im_no_1] is not None:
                            TETC[0][im_no_1] = np.zeros_like(tet_masks[start])
                        TETC[1][im_no_1] = tet_masks[im_no_1]
                    break
                else:
                    TETC[0][im_no] = I3B.copy()
                    TETC[1][im_no] = I2A.copy()
                    continue
            elif (ind == 0 or math.isnan(ind)) and ccel != 1:
                k += 1
                if k > thresh_next_cell:
                    break
                else:
                    continue

            k = 0
            pix = np.where(I2A == ind)
            pix0 = np.where(I2A != ind)
            
            # pix = np.flatnonzero(I2A == ind)
            # pix0 = np.flatnonzero(I2A != ind)
            
            I2A[pix] = ccel
            I2A[pix0] = 0
            
            IS6 = I2A.copy()
            I22 = np.zeros_like(I2)
            
            pix1 = np.where(IS6 == ccel)
            
            I2[pix1] = 0
            pix2 = np.unique(I2)
            pix2 = pix2[1:]  # Exclude background

            if ccel == 1:
                for ity, p2 in enumerate(pix2):
                    pix4 = np.where(I2 == p2)
                    I22[pix4] = ity + 1
                TETC[0][im_no] = IS6.copy()
            else:
                if len(pix2) > 0:
                    for ity, p2 in enumerate(pix2):
                        pix4 = np.where(I2 == p2)
                        I22[pix4] = ity + 1
                else:
                    I22 = I2.copy()
                IS61 = TETC[0][im_no]
                IS61[pix] = ccel
                TETC[0][im_no] = IS61.astype(np.uint16)

            TETC[1][im_no] = I22.copy()

        xx = -1
        for i in rang:
            if TETC[1][i] is not None and np.sum(TETC[1][i]) > 0:
                xx = i
                break
        ccel += 1
        rang2 = range(xx, len(tet_masks))
        #print(xx + 1)


    ccel -= 1  # number of cells tracked

    # Removing the shock-induced points from rang
    rang3 = list(rang)
    for start, end in [shock_period]:
        for i in range(start, end):
            if i in rang3:
                rang3.remove(i)

    all_obj = cal_allob1(ccel, TETC, rang)
    cell_data = cal_celldata(all_obj, ccel) ## double check values

    k = 1
    cell_artifacts = []
    for iv in range(ccel):
        if cell_data[iv, 2] < thresh_percent * len(rang3) or cell_data[iv, 4] > thresh:
            cell_artifacts.append(iv + 1)
            k += 1

    all_ccel = list(range(1, ccel + 1))

    if len(cell_artifacts) > 0:
        cell_artifacts = list(set(cell_artifacts))
        for iv in cell_artifacts:
            for its in rang3:
                pix = np.where(TETC[0][its] == iv)
                TETC[0][its][pix] = 0

    # Retaining and relabeling the new cells
    good_cells = sorted(set(all_ccel) - set(cell_artifacts))

    for iv in range(len(good_cells)):
        for its in rang3:
            pix = np.where(TETC[0][its] == good_cells[iv])
            TETC[0][its][pix] = iv + 1

    # Correcting the SpoSeg track masks or filling the empty spaces between the first and last appearance
    # Removing artifacts
    all_obj1 = cal_allob1(len(good_cells), TETC, rang)
    # plt.imshow(all_obj1, extent=[0, x_scale, 0, y_scale], aspect='auto',interpolation='nearest')
    cell_data1 = cal_celldata(all_obj1, len(good_cells))

    for iv in range(len(good_cells)):
        for its in range(int(cell_data1[iv, 0] + 1), int(cell_data1[iv, 1])):
            if all_obj1[iv, its] == 0:
                prev = np.where(all_obj1[iv, :its] > 0)[0][-1]
                all_obj1[iv, its] = all_obj1[iv, prev]
                pix = np.where(TETC[0][prev] == iv + 1)
                TETC[0][its][pix] = iv + 1

    # Cell array that contains the fully tracked TetSeg masks
    TETmasks = [TETC[0][i] for i in range(len(TETC[0]))]

    TET_obj = len(good_cells)
    all_obj_final = cal_allob2(TET_obj, TETmasks, list(range(len(TETmasks))))
    TET_Size = all_obj_final.copy()

    # Calculate first detection and last detection of tetrads
    TET_exists = np.zeros((2, TET_obj), dtype=int)
    for iv in range(TET_obj):
        TET_exists[0, iv] = np.where(TET_Size[iv, :] > 0)[0][0]  # 1st occurrence
        TET_exists[1, iv] = np.where(TET_Size[iv, :] > 0)[0][-1]  # last occurrence

    tet_masks_exists_tp = rang3
    TETmasks = replace_none_with_empty_array(TETmasks)
    return TET_obj, TET_exists, TETmasks

